from __future__ import absolute_import

import io
import time
import typing

from oslo_log import log
import pandas

import tobiko
from tobiko import config
from tobiko.tripleo import overcloud
from tobiko.shell import sh
from tobiko.openstack import topology


CONF = config.CONF
LOG = log.getLogger(__name__)


class PcsResourceException(tobiko.TobikoException):
    message = "pcs cluster is not in a healthy state"


def get_random_controller_ssh_client():
    """get a random controler's ssh client """
    nodes = topology.list_openstack_nodes(group='controller')
    controller_node = nodes[0]
    return controller_node.ssh_client


def get_pcs_resources_table(timeout=720, interval=2) -> pandas.DataFrame:
    """
    get pcs status from a controller and parse it
    to have it's resources states in check
       returns :
       rabbitmq-bundle-0    (ocf::heartbeat:rabbitmq-cluster):      Started con
       troller-0
     ip-10.0.0.101  (ocf::heartbeat:IPaddr2):       Started controller-1
       openstack-cinder-volume-docker-0     (ocf::heartbeat:docker):        Sta
       rted controller-0

    :return: dataframe of pcs resources stats table
    """
    failures: typing.List[str] = []
    start = time.time()

    ssh_client = get_random_controller_ssh_client()

    # prevent pcs table read failure while pacemaker is starting
    while time.time() - start < timeout:
        failures = []
        try:
            output = sh.execute("sudo pcs status resources |grep ocf",
                                ssh_client=ssh_client,
                                expect_exit_status=None).stdout
            # remove the first column when it only includes '*' characters
            output = output.replace('*', '').strip()
            stream = io.StringIO(output)
            table: pandas.DataFrame = pandas.read_csv(
                stream, delim_whitespace=True, header=None)
            table.columns = ['resource', 'resource_type', 'resource_state',
                             'overcloud_node']
        except ValueError:
            pcs_status_raw = sh.execute("sudo pcs status ",
                                        ssh_client=ssh_client,
                                        expect_exit_status=None).stdout
            failures.append(f'pcs status table import failed : '
                            f'pcs status stdout:\n {pcs_status_raw}')
            LOG.info('Retrying , timeout at: {}'
                     .format(timeout-(time.time() - start)))
            time.sleep(interval)
        else:
            break
    # exhausted all retries
    if failures:
        tobiko.fail('pcs status table import error\n' + '\n'.join(failures))

    LOG.debug("Got pcs status :\n%s", table)
    return table


def get_pcs_prefix_and_status_values():
    if topology.verify_osp_version('17.0', lower=True):
        ocf_prefix = "ocf::"
        promoted_status_str = "Master"
        unpromoted_status_str = "Slave"
    else:
        ocf_prefix = "ocf:"
        promoted_status_str = "Promoted"
        unpromoted_status_str = "Unpromoted"
    return ocf_prefix, promoted_status_str, unpromoted_status_str


class PacemakerResourcesStatus(object):
    """
    class to handle pcs resources checks
    """
    def __init__(self):
        self.pcs_df = get_pcs_resources_table()
        (self.ocf_prefix,
         self.promoted_status_str,
         self.unpromoted_status_str) = get_pcs_prefix_and_status_values()

    def container_runtime(self):

        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            f"({self.ocf_prefix}heartbeat:docker):")].empty:
            return 'docker'
        if not self.pcs_df[(self.pcs_df['resource_type'] ==
                            f"({self.ocf_prefix}heartbeat:podman):")].empty:
            return 'podman'

    def resource_count(self, resource_type):
        return self.pcs_df[(self.pcs_df['resource_type'] == resource_type)][
            'resource_state'].count()

    def resource_count_in_state(self, resource_type, resource_state):
        return self.pcs_df[(self.pcs_df['resource_type'] ==
                            resource_type) & (self.pcs_df['resource_state'] ==
                                              resource_state)][
            'resource_state'].count()

    def rabbitmq_resource_healthy(self):
        rabbitmq_resource_str = \
            f"({self.ocf_prefix}heartbeat:rabbitmq-cluster):"
        nodes_num = self.resource_count(rabbitmq_resource_str)
        started_num = self.resource_count_in_state(
            rabbitmq_resource_str, "Started")
        if nodes_num == started_num and nodes_num > 0:
            LOG.info("pcs status check: resource rabbitmq is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource rabbitmq not in healthy "
                     "state")
            return False

    def galera_resource_healthy(self):
        galera_resource_str = f"({self.ocf_prefix}heartbeat:galera):"
        nodes_num = self.resource_count(galera_resource_str)
        master_num = self.resource_count_in_state(
            galera_resource_str, self.promoted_status_str)
        if nodes_num == master_num and nodes_num > 0:
            LOG.info("pcs status check: resource galera is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource galera not in healthy state")
            return False

    def redis_resource_healthy(self):
        redis_resource_str = f"({self.ocf_prefix}heartbeat:redis):"
        if not overcloud.is_redis_expected():
            LOG.info("redis resource not expected on OSP 17 "
                     "and later releases by default")
            return self.pcs_df.query(
                f'resource_type == "{redis_resource_str}"').empty
        nodes_num = self.resource_count(redis_resource_str)
        master_num = self.resource_count_in_state(
            redis_resource_str, self.promoted_status_str)
        slave_num = self.resource_count_in_state(
            redis_resource_str, self.unpromoted_status_str)
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info("pcs status check: resource redis is in healthy state")
            return True
        else:
            LOG.info("pcs status check: resource redis not in healthy state")
            return False

    def vips_resource_healthy(self):
        if CONF.tobiko.tripleo.has_external_load_balancer:
            LOG.info("external load balancer used - "
                     "we can skip vips_resource sanity")
            return True
        else:
            vips_resource_str = f"({self.ocf_prefix}heartbeat:IPaddr2):"
            nodes_num = self.resource_count(vips_resource_str)
            started_num = self.resource_count_in_state(
                vips_resource_str, "Started")
            if nodes_num == started_num and nodes_num > 0:
                LOG.info("pcs status check: resources vips are "
                         "in healthy state")
                return True
            else:
                LOG.info(
                    "pcs status check: resources"
                    " vips are not in healthy state")
                return False

    def ha_proxy_cinder_healthy(self):
        if CONF.tobiko.tripleo.has_external_load_balancer:
            LOG.info("external load balancer used "
                     "- we can skip ha_proxy_resource sanity")
            return True
        else:
            ha_proxy_resource_str = (f"({self.ocf_prefix}heartbeat:"
                                     f"{self.container_runtime()}):")
            nodes_num = self.resource_count(ha_proxy_resource_str)
            started_num = self.resource_count_in_state(
                ha_proxy_resource_str, "Started")
            if nodes_num == started_num and nodes_num > 0:
                LOG.info("pcs status check: resources ha_proxy and"
                         " cinder are in healthy state")
                return True
            else:
                LOG.info(
                    "pcs status check: resources ha_proxy and cinder "
                    "are not in healthy state")
                return False

    def ovn_resource_healthy(self):
        ovn_resource_str = f"({self.ocf_prefix}ovn:ovndb-servers):"
        if self.pcs_df.query(
                f'resource_type == "{ovn_resource_str}"').empty:
            LOG.info('pcs status check: ovn is not deployed, skipping ovn '
                     'resource check')
            return True
        nodes_num = self.resource_count(ovn_resource_str)
        master_num = self.resource_count_in_state(
            ovn_resource_str, self.promoted_status_str)
        slave_num = self.resource_count_in_state(
            ovn_resource_str, self.unpromoted_status_str)
        if (master_num == 1) and (slave_num == nodes_num - master_num):
            LOG.info(
                "pcs status check: resource ovn is in healthy state")
            return True
        else:
            LOG.info(
                "pcs status check: resource ovn is in not in "
                "healthy state")
            return False

    @property
    def all_healthy(self):
        """
        check if each resource is in healthy order
        and return a global healthy status
        :return: Bool
        """
        for attempt_number in range(360):

            try:

                if all([
                   self.rabbitmq_resource_healthy(),
                   self.galera_resource_healthy(),
                   self.redis_resource_healthy(),
                   self.vips_resource_healthy(),
                   self.ha_proxy_cinder_healthy(),
                   self.ovn_resource_healthy()
                   ]):
                    LOG.info("pcs status checks: all resources are"
                             " in healthy state")
                    return True
                else:

                    LOG.info("pcs status check: not all resources are "
                             "in healthy "
                             "state")
                    raise PcsResourceException()
            except PcsResourceException:
                # reread pcs status
                LOG.info('Retrying pacemaker resource checks attempt '
                         '{} of 360'.format(attempt_number))
                time.sleep(1)
                self.pcs_df = get_pcs_resources_table()
        # exhausted all retries
        tobiko.fail('pcs cluster is not in a healthy state')


def get_overcloud_nodes_running_pcs_resource(resource=None,
                                             resource_type=None,
                                             resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    # pylint: disable=no-member
    pcs_df = get_pcs_resources_table()
    if resource:
        pcs_df_query_resource = pcs_df.query('resource=="{}"'.format(
                                        resource))
        return pcs_df_query_resource['overcloud_node'].unique().tolist()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'overcloud_node'].unique().tolist()

    if resource_type and not resource_state:
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['overcloud_node'].unique().tolist()


def get_resource_master_node(resource_type=None):
    get_overcloud_nodes_running_pcs_resource(
        resource_type=resource_type, resource_state='Master')


def get_ovn_db_master_node():
    ocf_prefix, promoted_status_str, _ = get_pcs_prefix_and_status_values()
    return get_overcloud_nodes_running_pcs_resource(
        resource_type=f'({ocf_prefix}ovn:ovndb-servers):',
        resource_state=promoted_status_str)


def get_overcloud_resource(resource_type=None,
                           resource_state=None):
    """
    Check what nodes are running the specified resource/type/state
    resource/type/state: exact str of a resource name as seen in pcs status
    :return: list of overcloud nodes
    """
    pcs_df = get_pcs_resources_table()

    if resource_type and resource_state:
        pcs_df_query_resource_type_state = pcs_df.query(
            'resource_type=="{}" and resource_state=="{}"'.format(
                resource_type, resource_state))
        return pcs_df_query_resource_type_state[
            'resource'].unique().tolist()

    if resource_type and not resource_state:
        # pylint: disable=no-member
        pcs_df_query_resource_type = pcs_df.query(
            'resource_type=="{}"'.format(resource_type))
        return pcs_df_query_resource_type['resource'].unique().tolist()


def instanceha_deployed():
    """check IHA deployment
    checks for existence of the nova-evacuate resource"""
    if overcloud.has_overcloud():
        return get_overcloud_nodes_running_pcs_resource(
            resource='nova-evacuate')
    else:
        return False


skip_if_instanceha_not_delpoyed = tobiko.skip_unless(
    'instanceha not delpoyed', instanceha_deployed)


def fencing_deployed():
    """check fencing deployment
    checks for existence of the stonith-fence type resources"""
    ssh_client = get_random_controller_ssh_client()
    fencing_output = sh.execute("sudo pcs status |grep "
                                "'stonith:fence_ipmilan'",
                                ssh_client=ssh_client,
                                expect_exit_status=None)

    if fencing_output.exit_status == 0:
        return True
    else:
        return False


skip_if_fencing_not_deployed = tobiko.skip_unless(
    'fencing not delpoyed', fencing_deployed)



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

import tobiko
from tobiko.openstack import octavia
from tobiko import config

LOG = log.getLogger(__name__)

CONF = config.CONF


def wait_for_status(status_key, status, get_client, object_id,
                    interval: tobiko.Seconds = None,
                    timeout: tobiko.Seconds = None,
                    error_ok=False, **kwargs):
    """Waits for an object to reach a specific status.

    :param status_key: The key of the status field in the response.
                       Ex. provisioning_status
    :param status: The status to wait for. Ex. "ACTIVE"
    :param get_client: The tobiko client get method.
                        Ex. _client.get_loadbalancer
    :param object_id: The id of the object to query.
    :param interval: How often to check the status, in seconds.
    :param timeout: The maximum time, in seconds, to check the status.
    :param error_ok: When true, ERROR status will not raise an exception.
    :raises TimeoutException: The object did not achieve the status or ERROR in
                              the check_timeout period.
    :raises UnexpectedStatusException: The request returned an unexpected
                                       response code.
    """

    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=(
                                        CONF.tobiko.octavia.check_timeout),
                                default_interval=(
                                        CONF.tobiko.octavia.check_interval)):
        response = get_client(object_id, **kwargs)
        if response[status_key] == status:
            return response

        if response[status_key] == octavia.ERROR and not error_ok:
            message = ('{name} {field} was updated to an invalid state of '
                       'ERROR'.format(name=get_client.__name__,
                                      field=status_key))
            raise octavia.RequestException(message)
        # it will raise tobiko.RetryTimeLimitError in case of timeout
        attempt.check_limits()

        LOG.debug(f"Waiting for {get_client.__name__} {status_key} to get "
                  f"from '{response[status_key]}' to '{status}'...")



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import os

import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import tests
from tobiko.openstack import stacks
from tobiko.shell import sh


@keystone.skip_unless_has_keystone_credentials()
class AgentTest(testtools.TestCase):

    def test_skip_if_missing_agents(self, count=1, should_skip=False,
                                    **params):
        if should_skip:
            expected_exception = self.skipException
        else:
            expected_exception = self.failureException

        @neutron.skip_if_missing_networking_agents(count=count, **params)
        def method():
            raise self.fail('Not skipped')

        exception = self.assertRaises(expected_exception, method)
        if should_skip:
            agents = neutron.list_agents(**params)
            message = "missing {!r} agent(s)".format(count - len(agents))
            if params:
                message += " with {!s}".format(
                    ','.join('{!s}={!r}'.format(k, v)
                             for k, v in params.items()))
            self.assertEqual(message, str(exception))
        else:
            self.assertEqual('Not skipped', str(exception))

    def test_skip_if_missing_agents_with_no_agents(self):
        self.test_skip_if_missing_agents(binary='never-never-land',
                                         should_skip=True)

    def test_skip_if_missing_agents_with_big_count(self):
        self.test_skip_if_missing_agents(count=1000000,
                                         should_skip=True)

    def test_neutron_agents_are_alive(self):
        agents = tests.test_neutron_agents_are_alive()
        # check has agents and they are all alive
        self.assertNotEqual([], agents)
        self.assertNotEqual([], agents.with_items(alive=True))

    def test_find_agents_with_binary(self):
        agent = neutron.list_agents().first
        agents = neutron.list_agents(binary=agent['binary'])
        self.assertIn(agent['id'], {a['id'] for a in agents})


class GetL3AgentModeTest(testtools.TestCase):

    @property
    def connection(self) -> sh.ShellConnection:
        return sh.local_shell_connection()

    def create_file(self, text: str) -> str:
        temp_dir = self.connection.make_temp_dir()
        temp_file = os.path.join(temp_dir, 'l3_agent.ini')
        with self.connection.open_file(temp_file, 'wt') as fd:
            fd.write(text)
        return temp_file

    def test_get_l3_agent_mode(self):
        text = """
        agent_mode=dvr
        """
        l3_agent_conf_path = self.create_file(text=text)
        result = neutron.get_l3_agent_mode(
            l3_agent_conf_path=l3_agent_conf_path,
            connection=self.connection)
        self.assertEqual('dvr', result)

    def test_get_l3_agent_mode_with_empty(self):
        l3_agent_conf_path = self.create_file(text='')
        result = neutron.get_l3_agent_mode(
            l3_agent_conf_path=l3_agent_conf_path,
            connection=self.connection)
        self.assertEqual('legacy', result)

    def test_get_l3_agent_mode_with_section(self):
        text = """
        [DEFAULT]
        agent_mode=dvr_no_external
        """
        l3_agent_conf_path = self.create_file(text=text)
        result = neutron.get_l3_agent_mode(
            l3_agent_conf_path=l3_agent_conf_path,
            connection=self.connection)
        self.assertEqual('dvr_no_external', result)

    def test_get_l3_agent_mode_with_wrong_section(self):
        text = """
        [wrong-section]
        agent_mode=dvr
        """
        l3_agent_conf_path = self.create_file(text=text)
        result = neutron.get_l3_agent_mode(
            l3_agent_conf_path=l3_agent_conf_path,
            connection=self.connection)
        self.assertEqual('legacy', result)

    def test_get_l3_agent_mode_with_tricky_option(self):
        text = """
        agent_mode_2=dvr
        """
        l3_agent_conf_path = self.create_file(text=text)
        result = neutron.get_l3_agent_mode(
            l3_agent_conf_path=l3_agent_conf_path,
            connection=self.connection)
        self.assertEqual('legacy', result)


class SSHGetL3AgentModeTest(GetL3AgentModeTest):

    server_stack = tobiko.required_fixture(stacks.UbuntuServerStackFixture)

    @property
    def connection(self) -> sh.SSHShellConnection:
        connection = sh.shell_connection()
        if isinstance(connection, sh.SSHShellConnection):
            return connection
        return self.server_stack.connection



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from tobiko.openstack.octavia import _client

LoadBalancerType = typing.Dict[str, typing.Any]
LoadBalancerIdType = typing.Union[str, LoadBalancerType]


def get_load_balancer_id(load_balancer: LoadBalancerIdType) -> str:
    if isinstance(load_balancer, str):
        return load_balancer
    else:
        return load_balancer['id']


def get_load_balancer(load_balancer: LoadBalancerIdType,
                      client: _client.OctaviaClientType = None) \
        -> LoadBalancerType:
    load_balancer_id = get_load_balancer_id(load_balancer)
    return _client.octavia_client(client).load_balancer_show(
        load_balancer_id)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import testtools

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import neutron
from tobiko.openstack import stacks


LOG = log.getLogger(__name__)


@keystone.skip_unless_has_keystone_credentials()
class NetworkTest(testtools.TestCase):
    """Tests network creation"""

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.NetworkStackFixture)

    @neutron.skip_if_missing_networking_extensions('port-security')
    def test_port_security_enabled(self):
        self.assertEqual(self.stack.port_security_enabled,
                         self.stack.network_details['port_security_enabled'])
        self.assertEqual(self.stack.port_security_enabled,
                         self.stack.outputs.port_security_enabled)

    @neutron.skip_if_missing_networking_extensions('net-mtu')
    def test_net_mtu(self):
        self.assertEqual(self.stack.network_details['mtu'],
                         self.stack.outputs.mtu)

    def test_ipv4_subnet_cidr(self):
        if not self.stack.has_ipv4:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no ipv4 "
                             "subnet")

        subnet = neutron.find_subnet(cidr=str(self.stack.ipv4_subnet_cidr))
        self.assertEqual(neutron.get_subnet(self.stack.ipv4_subnet_id), subnet)

    def test_ipv6_subnet_cidr(self):
        if not self.stack.has_ipv6:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no ipv6 "
                             "subnet")
        subnet = neutron.find_subnet(cidr=str(self.stack.ipv6_subnet_cidr))
        self.assertEqual(neutron.get_subnet(self.stack.ipv6_subnet_id), subnet)

    def test_gateway_network(self):
        if not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no gateway")
        self.assertEqual(
            self.stack.gateway_network_id,
            self.stack.gateway_details['external_gateway_info']['network_id'])

    def test_ipv4_subnet_gateway_ip(self):
        if not self.stack.has_ipv4 or not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no IPv4 "
                             "gateway")
        self.assertIn(
            self.stack.ipv4_subnet_gateway_ip,
            self.stack.ipv4_gateway_addresses)

    def test_ipv6_subnet_gateway_ip(self):
        if not self.stack.has_ipv6 or not self.stack.has_gateway:
            tobiko.skip_test(f"Stack {self.stack.stack_name} has no IPv6 "
                             "gateway")
        self.assertIn(
            self.stack.ipv6_subnet_gateway_ip,
            self.stack.ipv6_gateway_addresses)


@keystone.skip_unless_has_keystone_credentials()
@neutron.skip_if_missing_networking_extensions('net-mtu-write')
class NetworkWithNetMtuWriteTest(NetworkTest):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(
        stacks.NetworkWithNetMtuWriteStackFixture)

    def test_net_mtu_write(self):
        self.assertEqual(self.stack.mtu, self.stack.outputs.mtu)


@keystone.skip_unless_has_keystone_credentials()
@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(neutron.L3_AGENT, 2)
class L3HaNetworkTest(NetworkTest):

    #: Stack of resources with a network with a gateway router
    stack = tobiko.required_fixture(stacks.L3haNetworkStackFixture)


@keystone.skip_unless_has_keystone_credentials()
@stacks.skip_unless_has_external_network
class ExternalNetworkTest(testtools.TestCase):

    def test_get_external_network(self):
        network = stacks.get_external_network()
        self.assertTrue(network['id'])
        self.assertIs(True, network['router:external'])
        self.assertEqual('ACTIVE', network['status'])
        subnets = neutron.list_subnets(network_id=network['id'])
        self.assertNotEqual([], subnets)

    def test_has_external_network(self):
        self.assertIs(True, stacks.has_external_network())


@keystone.skip_unless_has_keystone_credentials()
@stacks.skip_unless_has_floating_network
class RouterTest(testtools.TestCase):

    def test_get_floating_network(self):
        network = stacks.get_floating_network()
        self.assertTrue(network['id'])
        self.assertIs(True, network['router:external'])
        self.assertEqual('ACTIVE', network['status'])

    def test_has_floating_network(self):
        self.assertTrue(stacks.has_floating_network())

    def test_get_router(self):
        router = stacks.get_router()
        self.assertIsNotNone(router)
        self.assertEqual(
            stacks.get_floating_network_id(),
            router['external_gateway_info']['network_id'])

    def test_has_router(self):
        self.assertTrue(stacks.has_router())


class RouterInterfaceTestRouter(stacks.RouterStackFixture):
    pass


class RouterInterfaceTestNetwork(stacks.NetworkStackFixture):
    pass


@keystone.skip_unless_has_keystone_credentials()
@stacks.skip_unless_has_floating_network
class RouterInterfaceTest(testtools.TestCase):

    router_stack = tobiko.required_fixture(RouterInterfaceTestRouter)
    network_stack = tobiko.required_fixture(RouterInterfaceTestNetwork)

    required_fixtures = [router_stack, network_stack]

    @classmethod
    def tearDownClass(cls) -> None:
        for fixture in cls.required_fixtures:
            try:
                tobiko.cleanup_fixture(fixture.fixture)
            except Exception:
                LOG.exception(f'Error cleaning up fixture: {fixture.fixture}')

    def test_ensure_router_interface_with_subnet(self,
                                                 ip_version=4):
        network = neutron.create_network()
        subnet = neutron.create_subnet(network=network,
                                       ip_version=ip_version)
        self._test_ensure_router(subnet=subnet)

    def test_ensure_router_interface_with_ipv6_subnet(self):
        self.test_ensure_router_interface_with_subnet(ip_version=6)

    def test_ensure_router_interface_with_routed_ipv4_subnet(self):
        self._test_ensure_router(subnet=self.network_stack.ipv4_subnet_id)

    def test_ensure_router_interface_with_routed_ipv6_subnet(self):
        self._test_ensure_router(subnet=self.network_stack.ipv6_subnet_id)

    def test_ensure_router_interface_with_network(self):
        network = neutron.create_network()
        neutron.create_subnet(network=network)
        self._test_ensure_router(network=network)

    def test_ensure_router_interface_with_routed_network(self):
        self._test_ensure_router(network=self.network_stack.network_id)

    def _test_ensure_router(self,
                            network: neutron.NetworkIdType = None,
                            subnet: neutron.SubnetIdType = None):
        router = self.router_stack.router_details
        self.assertRaises(tobiko.ObjectNotFound,
                          neutron.find_port,
                          device=router,
                          subnet=subnet,
                          network=network)
        port = stacks.ensure_router_interface(router=router,
                                              network=network,
                                              subnet=subnet,
                                              add_cleanup=True)
        self.assertEqual(router['id'], port['device_id'])
        self.assertEqual(neutron.find_port(device=router,
                                           network=network,
                                           subnet=subnet)['id'],
                         port['id'])
        self.assertEqual(port['id'],
                         stacks.ensure_router_interface(
                             router=router,
                             network=network,
                             subnet=subnet,
                             add_cleanup=True)['id'])
        return port



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import netaddr

import tobiko
from tobiko.openstack import heat
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.shiftstack import _heat
from tobiko.shiftstack import _neutron


class ShiftstackHeatStackFixture(heat.HeatStackFixture):

    def __init__(self,
                 neutron_client: neutron.NeutronClientType = None,
                 **params):
        super().__init__(**params)
        self._neutron_client = neutron_client

    def setup_client(self) -> heat.HeatClient:
        if self.client is None:
            self.client = _heat.get_shiftstack_heat_client()
        return self.client

    @property
    def neutron_client(self) -> neutron.NeutronClientType:
        if self._neutron_client is None:
            self._neutron_client = _neutron.get_shiftstack_neutron_client()
        return self._neutron_client


class ShiftstackRouterStackFixture(ShiftstackHeatStackFixture,
                                   stacks.RouterStackFixture):

    @staticmethod
    def create_router_interface(
            router: neutron.RouterIdType = None,
            subnet: neutron.SubnetIdType = None,
            network: neutron.NetworkIdType = None,
            client: neutron.NeutronClientType = None,
            add_cleanup=False) -> neutron.PortType:
        stack = ShiftstackRouterInterfaceStackFixture(
            router=router,
            subnet=subnet,
            network=network,
            neutron_client=client)
        if add_cleanup:
            tobiko.use_fixture(stack)
        else:
            tobiko.setup_fixture(stack)
        return stack.port_details


class ShiftstackRouterInterfaceStackFixture(
        ShiftstackHeatStackFixture, stacks.RouterInterfaceStackFixture):
    pass


class ShiftstackFloatingIpStackFixture(ShiftstackHeatStackFixture,
                                       stacks.FloatingIpStackFixture):
    router_stack = tobiko.required_fixture(ShiftstackRouterStackFixture)


def ensure_shiftstack_node_floating_ip(
        server: nova.ServerType,
        client: neutron.NeutronClientType = None) \
        -> netaddr.IPAddress:
    client = _neutron.shiftstack_neutron_client(client)
    fixed_ip_address = _neutron.find_shiftstack_node_ip_address(
        server=server, client=client)
    for floating_ip in neutron.list_floating_ips(client=client):
        if (netaddr.IPAddress(floating_ip['fixed_ip_address']) ==
                fixed_ip_address):
            break
    else:
        fixture = ShiftstackFloatingIpStackFixture(
            device_id=nova.get_server_id(server),
            fixed_ip_address=str(fixed_ip_address))
        floating_ip = tobiko.setup_fixture(fixture).floating_ip_details
    return netaddr.IPAddress(floating_ip['floating_ip_address'])



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import time

import mock

import tobiko
from tobiko import common


class PatchMixin(object):
    """Mixin class with mock method helpers"""

    def patch(self, obj, attribute, value=mock.DEFAULT, spec=None,
              create=False, spec_set=None, autospec=None,
              new_callable=None, **kwargs):
        # pylint: disable=arguments-differ
        context = mock.patch.object(target=obj, attribute=attribute, new=value,
                                    spec=spec, create=create,
                                    spec_set=spec_set, autospec=autospec,
                                    new_callable=new_callable, **kwargs)
        mocked = context.start()
        self.addCleanup(context.stop)
        return mocked

    def patch_time(self, current_time=None, time_increment=None):
        if not hasattr(self, 'mock_time'):
            self.mock_time = PatchTimeFixture(current_time=current_time,
                                              time_increment=time_increment)
        else:
            self.mock_time.patch_time(current_time=current_time,
                                      time_increment=time_increment)
        return self.useFixture(self.mock_time)


class PatchFixture(PatchMixin, tobiko.SharedFixture):
    """Fixture class with mock method helpers"""


class PatchTimeFixture(PatchFixture):

    start_time = 0.
    current_time = 0.
    time_increment = 1.

    def __init__(self, current_time=None, time_increment=.1):
        self.time = mock.MagicMock(specs=time.time, side_effect=self._time)
        self.sleep = mock.MagicMock(specs=time.sleep, side_effect=self._sleep)
        self.patch_time(current_time=current_time,
                        time_increment=time_increment)

    def setup_fixture(self):
        # pylint: disable=protected-access
        self.patch(common._time, '_time', self)

    def _time(self):
        result = self.current_time
        self.current_time += self.time_increment
        return result

    def _sleep(self, seconds):
        self.current_time += seconds

    def patch_time(self, current_time=0., time_increment=.1):
        if current_time is not None:
            self.start_time = self.current_time = current_time
        if time_increment is not None:
            self.time_increment = time_increment



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing  # noqa
from os.path import expanduser

from oslo_log import log

import tobiko


LOG = log.getLogger(__name__)


def get_user_home_dir():
    """return a str path of the current user's home"""
    return expanduser("~")


class ExecutePathFixture(tobiko.SharedFixture):

    def __init__(self, executable_dirs=None, environ=None):
        super(ExecutePathFixture, self).__init__()
        self.executable_dirs = list(executable_dirs or
                                    [])  # type: typing.List[str]
        self.environ = dict(environ or {})  # type: typing.Dict[str, str]

    def setup_fixture(self):
        missing_dirs = []
        for executable_dir in self.executable_dirs:
            if (executable_dir not in self.path_dirs and
                    executable_dir not in missing_dirs):
                missing_dirs.append(executable_dir)
        if missing_dirs:
            new_path_dirs = missing_dirs + self.path_dirs
            self.addCleanup(self.remove_executable_dirs, missing_dirs)
            self.environ['PATH'] = ':'.join(new_path_dirs)
            LOG.debug('Directories added to executable path: %s',
                      ', '.join(sorted(new_path_dirs)))

    @property
    def path_dirs(self):
        return list(self.environ.get('PATH', '').split(':'))

    def remove_executable_dirs(self, removal_dirs):
        removal_dirs = set(removal_dirs) & set(self.path_dirs)
        if removal_dirs:
            new_path_dirs = [p
                             for p in self.path_dirs
                             if p not in removal_dirs]
            self.environ['PATH'] = ':'.join(new_path_dirs)
            LOG.debug('Directories removed from executable path: %s',
                      ', '.join(sorted(removal_dirs)))



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools

from tobiko.shell import sh


class NameserversTest(testtools.TestCase):

    ssh_client = None

    def test_list_nameservers(self):
        nameservers = sh.list_nameservers(ssh_client=self.ssh_client)
        self.assertNotEqual([], nameservers)
        return nameservers

    def test_list_nameservers_with_ip_version_4(self):
        nameservers = sh.list_nameservers(ssh_client=self.ssh_client,
                                          ip_version=4)
        for nameserver in nameservers:
            self.assertEqual(4, nameserver.version, nameserver)

    def test_list_nameservers_with_ip_version_6(self):
        nameservers = sh.list_nameservers(ssh_client=self.ssh_client,
                                          ip_version=6)
        for nameserver in nameservers:
            self.assertEqual(6, nameserver.version, nameserver)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import getpass
import io
import os.path
import shutil
import socket
import tempfile
import typing

import paramiko
from oslo_log import log

import tobiko
from tobiko.shell.sh import _command
from tobiko.shell.sh import _exception
from tobiko.shell.sh import _execute
from tobiko.shell.sh import _hostname
from tobiko.shell.sh import _mktemp
from tobiko.shell import ssh


LOG = log.getLogger(__name__)

ShellConnectionType = typing.Union['ShellConnection', ssh.SSHClientType, str]


def connection_hostname(connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).hostname


def connection_login(connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).login


def connection_username(connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).username


def is_local_connection(connection: ShellConnectionType = None) -> bool:
    return shell_connection(connection).is_local


def is_cirros_connection(connection: ShellConnectionType = None) -> bool:
    return shell_connection(connection).is_cirros


def get_file(remote_file: str,
             local_file: str,
             connection: ShellConnectionType = None):
    return shell_connection(connection).get_file(
        remote_file=remote_file, local_file=local_file)


def open_file(filename: typing.Union[str, bytes],
              mode: str,
              buffering: int = None,
              connection: ShellConnectionType = None,
              sudo=False) \
            -> typing.Union[typing.IO, paramiko.sftp_file.SFTPFile]:
    connection = shell_connection(connection)
    if sudo:
        if 'r' not in mode:
            raise ValueError('sudo only supported in reading mode')
        temp_file = connection.make_temp_file(auto_clean=True)
        connection.execute(['cp', filename, temp_file], sudo=True)
        filename = temp_file
    return connection.open_file(filename=filename,
                                mode=mode,
                                buffering=buffering)


def put_file(local_file: str,
             remote_file: str,
             connection: ShellConnectionType = None):
    return shell_connection(connection).put_file(
        local_file=local_file, remote_file=remote_file)


def put_files(*local_files: str,
              remote_dir: str,
              connection: ShellConnectionType = None):
    return shell_connection(connection).put_files(*local_files,
                                                  remote_dir=remote_dir)


def make_temp_file(auto_clean=True,
                   connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).make_temp_file(
        auto_clean=auto_clean)


def make_temp_dir(auto_clean=True,
                  sudo: bool = None,
                  connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).make_temp_dir(
        auto_clean=auto_clean, sudo=sudo)


def remove_files(filename: str, *filenames: str,
                 connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).remove_files(
        filename, *filenames)


def make_dirs(name: str,
              exist_ok=True,
              connection: ShellConnectionType = None) -> str:
    return shell_connection(connection).make_dirs(
        name=name, exist_ok=exist_ok)


def local_shell_connection() -> 'LocalShellConnection':
    return tobiko.get_fixture(LocalShellConnection)


def shell_connection(obj: ShellConnectionType = None,
                     manager: 'ShellConnectionManager' = None) \
        -> 'ShellConnection':
    if isinstance(obj, ShellConnection):
        return obj
    else:
        return shell_connection_manager(manager).get_shell_connection(obj)


def ssh_shell_connection(obj: ShellConnectionType = None,
                         manager: 'ShellConnectionManager' = None) \
        -> 'SSHShellConnection':
    connection = shell_connection(obj, manager=manager)
    return tobiko.check_valid_type(connection, SSHShellConnection)


def register_shell_connection(connection: 'ShellConnection',
                              manager: 'ShellConnectionManager' = None) -> \
        None:
    tobiko.check_valid_type(connection, ShellConnection)
    shell_connection_manager(manager).register_shell_connection(connection)


def shell_connection_manager(manager: 'ShellConnectionManager' = None):
    if manager is None:
        return tobiko.setup_fixture(ShellConnectionManager)
    else:
        tobiko.check_valid_type(manager, ShellConnectionManager)
        return manager


ShellConnectionKey = typing.Optional[ssh.SSHClientFixture]


class ShellConnectionManager(tobiko.SharedFixture):

    def __init__(self,
                 local_hostnames: typing.Iterable[str] = None):
        super(ShellConnectionManager, self).__init__()
        self._host_connections: typing.Dict['ShellConnectionKey',
                                            'ShellConnection'] = {}
        if local_hostnames is not None:
            local_hostnames = set(local_hostnames)
        self._local_hostnames = local_hostnames

    @property
    def local_hostnames(self) -> typing.Set[str]:
        if self._local_hostnames is None:
            hostname = socket.gethostname()
            self._local_hostnames = {'localhost',
                                     tobiko.get_short_hostname(hostname)}
        return self._local_hostnames

    def get_shell_connection(self,
                             obj: typing.Union[str, ssh.SSHClientType]) -> \
            'ShellConnection':
        ssh_client: typing.Optional[ssh.SSHClientFixture]
        if isinstance(obj, str):
            if obj in self.local_hostnames:
                ssh_client = None  # local connection
            else:
                ssh_client = ssh.ssh_client(host=obj)
        else:
            ssh_client = ssh.ssh_client_fixture(obj)
        connection = self._host_connections.get(ssh_client)
        if connection is None:
            connection = self._get_shell_connection(ssh_client=ssh_client)
            self._host_connections[ssh_client] = connection
        return connection

    def register_shell_connection(self, connection: 'ShellConnection'):
        ssh_client = ssh.ssh_client_fixture(connection.ssh_client)
        self._host_connections[ssh_client] = connection

    @staticmethod
    def _get_shell_connection(ssh_client: ssh.SSHClientFixture = None) \
            -> 'ShellConnection':
        if ssh_client is None:
            return local_shell_connection()
        else:
            return SSHShellConnection(ssh_client=ssh_client)


class ShellConnection(tobiko.SharedFixture):

    def connect(self) -> 'ShellConnection':
        return tobiko.setup_fixture(self)

    def close(self) -> 'ShellConnection':
        return tobiko.cleanup_fixture(self)

    def reconnect(self):
        return tobiko.reset_fixture(self)

    @property
    def hostname(self) -> str:
        raise NotImplementedError

    @property
    def ssh_client(self) -> ssh.SSHClientType:
        raise NotImplementedError

    @property
    def is_local(self) -> bool:
        raise NotImplementedError

    @property
    def is_cirros(self) -> bool:
        return False

    @property
    def username(self) -> str:
        raise NotImplementedError

    @property
    def login(self) -> str:
        return f"{self.username}@{self.hostname}"

    def get_config_path(self, path: str) -> str:
        raise NotImplementedError

    def execute(self,
                command: _command.ShellCommandType,
                *args, **execute_params) -> \
            _execute.ShellExecuteResult:
        execute_params.setdefault('ssh_client', self.ssh_client)
        return _execute.execute(command, *args, **execute_params)

    def put_file(self, local_file: str, remote_file: str):
        raise NotImplementedError

    def put_files(self,
                  *local_files: str,
                  remote_dir: str,
                  make_dirs=True):
        # pylint: disable=redefined-outer-name
        remote_dir = os.path.normpath(remote_dir)
        put_files = {}
        for local_file in local_files:
            local_file = os.path.normpath(local_file)
            if os.path.isdir(local_file):
                top_dir = os.path.dirname(local_file)
                for local_dir, _, files in os.walk(local_file):
                    for filename in files:
                        local_file = os.path.join(local_dir, filename)
                        remote_file = os.path.join(
                            remote_dir,
                            os.path.relpath(local_file, start=top_dir))
                        put_files[os.path.realpath(local_file)] = remote_file
            else:
                remote_file = os.path.join(
                    remote_dir, os.path.basename(local_file))
                put_files[os.path.realpath(local_file)] = remote_file
        remote_dirs = set()
        for local_file, remote_file in sorted(put_files.items()):
            if make_dirs:
                remote_dir = os.path.dirname(remote_file)
                if remote_dir not in remote_dirs:
                    self.make_dirs(remote_dir, exist_ok=True)
                    remote_dirs.add(remote_dir)
            self.put_file(local_file, remote_file)

    def get_file(self, remote_file: str, local_file: str):
        raise NotImplementedError

    def get_environ(self) -> typing.Dict[str, str]:
        raise NotImplementedError

    def open_file(self,
                  filename: typing.Union[str, bytes],
                  mode: str,
                  buffering: int = None) \
            -> typing.Union[typing.IO, paramiko.sftp_file.SFTPFile]:
        raise NotImplementedError

    def __str__(self) -> str:
        return f"{type(self).__name__}<{self.login}>"

    def exists(self, path: str) -> bool:
        raise NotImplementedError

    def is_file(self, path: str) -> bool:
        raise NotImplementedError

    def is_directory(self, path: str) -> bool:
        raise NotImplementedError

    def make_temp_file(self, auto_clean=True) -> str:
        raise NotImplementedError

    def make_temp_dir(self, auto_clean=True, sudo: bool = None) -> str:
        raise NotImplementedError

    def remove_files(self, filename: str, *filenames: str):
        raise NotImplementedError

    def make_dirs(self, name: str, exist_ok=True):
        raise NotImplementedError


class LocalShellConnection(ShellConnection):

    @property
    def ssh_client(self) -> bool:
        return False

    @property
    def is_local(self) -> bool:
        return True

    @property
    def username(self) -> str:
        return getpass.getuser()

    @property
    def hostname(self) -> str:
        return socket.gethostname()

    def get_environ(self) -> typing.Dict[str, str]:
        return dict(os.environ)

    def put_file(self, local_file: str, remote_file: str):
        LOG.debug(f"Copy local file as {self.login}: '{local_file}' -> "
                  f"'{remote_file}' ...")
        shutil.copyfile(local_file, remote_file)

    def get_file(self, remote_file: str, local_file: str):
        LOG.debug(f"Copy local file as {self.login}: '{remote_file}' -> "
                  f"'{local_file}' ...")
        if local_file.startswith('~'):
            local_file = os.path.expanduser(local_file)
        if remote_file.startswith('~'):
            remote_file = os.path.expanduser(remote_file)
        shutil.copyfile(remote_file, local_file)

    def open_file(self,
                  filename: typing.Union[str, bytes],
                  mode: str,
                  buffering: int = None) -> typing.IO:
        params: typing.Dict[str, typing.Any] = {}
        if buffering is not None:
            params.update(buffering=buffering)
        return io.open(file=filename, mode=mode, **params)

    def exists(self, path: str) -> bool:
        return os.path.exists(path)

    def is_file(self, path: str) -> bool:
        return os.path.isfile(path)

    def is_directory(self, path: str) -> bool:
        return os.path.isdir(path)

    def make_temp_file(self, auto_clean=True) -> str:
        fd, temp_file = tempfile.mkstemp()
        os.close(fd)
        if auto_clean:
            tobiko.add_cleanup(self.remove_files, temp_file)
        return temp_file

    def make_temp_dir(self, auto_clean=True, sudo: bool = None) -> str:
        if sudo:
            return _mktemp.make_temp_dir(ssh_client=self.ssh_client,
                                         sudo=True)
        else:
            temp_dir = tempfile.mkdtemp()
            LOG.debug(f"Local temporary directory created as {self.login}: "
                      f"{temp_dir}")
            if auto_clean:
                tobiko.add_cleanup(self.remove_files, temp_dir)
            return temp_dir

    def remove_files(self, filename: str, *filenames: str):
        filenames = (filename,) + filenames
        LOG.debug(f"Remove local files as {self.login}: {filenames}")
        for filename in filenames:
            if os.path.isdir(filename):
                shutil.rmtree(filename)
            elif os.path.exists(filename):
                os.remove(filename)

    def make_dirs(self, name: str, exist_ok=True):
        os.makedirs(name=name,
                    exist_ok=exist_ok)

    def get_config_path(self, path: str) -> str:
        return tobiko.tobiko_config_path(path)


class SSHShellConnection(ShellConnection):

    def __init__(self,
                 ssh_client: ssh.SSHClientFixture = None):
        super().__init__()
        self._ssh_client = ssh_client

    def setup_fixture(self):
        self._ssh_client.connect()

    def cleanup_fixture(self):
        self._ssh_client.close()

    @property
    def ssh_client(self) -> ssh.SSHClientFixture:
        if self._ssh_client is None:
            raise ValueError('Unspecified SSH client')
        return self._ssh_client

    @property
    def is_local(self) -> bool:
        return False

    @property
    def username(self) -> str:
        return self.ssh_client.setup_connect_parameters()['username']

    _hostname: typing.Optional[str] = None

    @property
    def hostname(self) -> str:
        if self._hostname is None:
            self._hostname = _hostname.ssh_hostname(ssh_client=self.ssh_client)
        return self._hostname

    _sftp: typing.Optional[paramiko.SFTPClient] = None

    @property
    def sftp_client(self) -> paramiko.SFTPClient:
        if self._sftp is None:
            self._sftp = self.ssh_client.connect().open_sftp()
        return self._sftp

    def get_environ(self) -> typing.Dict[str, str]:
        lines = self.execute('source /etc/profile; env').stdout.splitlines()
        return dict(_parse_env_line(line)
                    for line in lines
                    if line.strip())

    def put_file(self, local_file: str, remote_file: str):
        LOG.debug(f"Put remote file as {self.login}: '{local_file}' -> "
                  f"'{remote_file}'...")
        self.sftp_client.put(local_file, remote_file)

    def open_file(self,
                  filename: typing.Union[str, bytes],
                  mode: str,
                  buffering: int = None) -> paramiko.sftp_file.SFTPFile:
        if buffering is None:
            buffering = -1
        return self.sftp_client.open(filename=filename,
                                     mode=mode,
                                     bufsize=buffering)

    def get_file(self, remote_file: str, local_file: str):
        LOG.debug(f"Get remote file as {self.login}: '{remote_file}' -> "
                  f"'{local_file}'...")
        if local_file.startswith('~'):
            local_file = os.path.expanduser(local_file)
        if remote_file.startswith('~'):
            remote_file = self.execute(f'echo {remote_file}').stdout.strip()
        self.sftp_client.get(remote_file, local_file)

    def exists(self, path: str) -> bool:
        try:
            self.execute(['test', '-e', path])
        except _exception.ShellCommandFailed:
            return False
        else:
            return True

    def is_file(self, path: str) -> bool:
        try:
            self.execute(['test', '-f', path])
        except _exception.ShellCommandFailed:
            return False
        else:
            return True

    def is_directory(self, path: str) -> bool:
        try:
            self.execute(['test', '-d', path])
        except _exception.ShellCommandFailed:
            return False
        else:
            return True

    def make_temp_file(self, auto_clean=True) -> str:
        temp_file = self.execute('mktemp').stdout.strip()
        LOG.debug(f"Remote temporary file created as {self.login}: "
                  f"{temp_file}")
        if auto_clean:
            tobiko.add_cleanup(self.remove_files, temp_file)
        return temp_file

    def make_temp_dir(self, auto_clean=True, sudo: bool = None) -> str:
        temp_dir = self.execute('mktemp -d', sudo=sudo).stdout.strip()
        LOG.debug(f"Remote temporary directory created as {self.login}: "
                  f"{temp_dir}")
        if auto_clean:
            tobiko.add_cleanup(self.remove_files, temp_dir)
        return temp_dir

    def remove_files(self, filename: str, *filenames: str):
        filenames = (filename,) + filenames
        LOG.debug(f"Remove remote files as {self.login}: {filenames}")
        command = _command.shell_command('rm -fR') + filenames
        self.execute(command)

    def make_dirs(self, name: str, exist_ok=True):
        command = _command.shell_command('mkdir')
        if exist_ok:
            command += '-p'
        command += name
        self.execute(command)

    _user_dir: typing.Optional[str] = None

    @property
    def user_dir(self) -> str:
        if self._user_dir is None:
            self._user_dir = self.execute('sh -c "echo $HOME"').stdout.strip()
        return self._user_dir

    def get_config_path(self, path: str) -> str:
        if path[0] in '~.':
            path = f"{self.user_dir}{path[1:]}"
        return path


def _parse_env_line(line: str) -> typing.Tuple[str, str]:
    name, value = line.split('=', 1)
    return name.strip(), value.strip()



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import contextlib
import socket
import urllib

from oslo_log import log
import sshtunnel

import tobiko
from tobiko.shell.ssh import _client


LOG = log.getLogger(__name__)


def get_forward_port_address(address, ssh_client=None, manager=None):
    if ssh_client is None:
        ssh_client = _client.ssh_proxy_client()
    manager = manager or DEFAULT_SSH_PORT_FORWARD_MANAGER
    return manager.get_forward_port_address(address, ssh_client=ssh_client)


def get_forward_url(url, ssh_client=None, manager=None):
    url = parse_url(url)
    if ssh_client is None:
        ssh_client = _client.ssh_proxy_client()
    manager = manager or DEFAULT_SSH_PORT_FORWARD_MANAGER
    address = binding_address(url)
    forward_address = get_forward_port_address(address, ssh_client=ssh_client,
                                               manager=manager)
    return binding_url(forward_address)


def reset_default_ssh_port_forward_manager():
    # pylint: disable=global-statement
    global DEFAULT_SSH_PORT_FORWARD_MANAGER
    DEFAULT_SSH_PORT_FORWARD_MANAGER = SSHPortForwardManager()


class SSHPortForwardManager(object):

    def __init__(self):
        self.forward_addresses = {}
        self.forwarders = {}

    def get_forward_port_address(self, address, ssh_client):
        try:
            return self.forward_addresses[address, ssh_client]
        except KeyError:
            pass

        forwarder = self.get_forwarder(address, ssh_client=ssh_client)
        if forwarder:
            forward_address = forwarder.get_forwarding(address)
        else:
            forward_address = address

        self.forward_addresses[address, ssh_client] = forward_address
        return forward_address

    def get_forwarder(self, address, ssh_client):
        try:
            return self.forwarders[address, ssh_client]
        except KeyError:
            pass

        if ssh_client:
            tobiko.check_valid_type(ssh_client, _client.SSHClientFixture)
            forwarder = SSHTunnelForwarderFixture(ssh_client=ssh_client)
            forwarder.put_forwarding(address)
            tobiko.setup_fixture(forwarder)
        else:
            forwarder = None

        self.forwarders[address, ssh_client] = forwarder
        return forwarder


DEFAULT_SSH_PORT_FORWARD_MANAGER = SSHPortForwardManager()


class SSHTunnelForwarderFixture(tobiko.SharedFixture):

    forwarder = None

    def __init__(self, ssh_client):
        super(SSHTunnelForwarderFixture, self).__init__()
        self.ssh_client = ssh_client
        self._forwarding = collections.OrderedDict()

    def put_forwarding(self, remote, local=None):
        if not local:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            with contextlib.closing(sock):
                sock.bind(('127.0.0.1', 0))
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                hostname, port = sock.getsockname()
                local = hostname, port
        return self._forwarding.setdefault(remote, local)

    def get_forwarding(self, remote):
        return self._forwarding.get(remote)

    def setup_fixture(self):
        self.setup_forwarder()

    def setup_forwarder(self):
        forwarder = self.forwarder
        if not forwarder:
            remote_bind_addresses = list(self._forwarding.keys())
            local_bind_addresses = list(self._forwarding.values())
            self.forwarder = forwarder = SSHTunnelForwarder(
                ssh_client=self.ssh_client,
                local_bind_addresses=local_bind_addresses,
                remote_bind_addresses=remote_bind_addresses)
            self.addCleanup(self.cleanup_forwarder)
            forwarder.start()
            self.ssh_client.addCleanup(self.cleanup_forwarder)
        return forwarder

    def cleanup_forwarder(self):
        forwarder = self.forwarder
        if forwarder:
            del self.forwarder
            forwarder.stop()


# pylint: disable=protected-access
class SSHUnixForwardHandler(sshtunnel._ForwardHandler):

    transport = None

    def handle(self):
        uid = sshtunnel.get_connection_id()
        self.info = '#{0} <-- {1}'.format(uid, self.client_address or
                                          self.server.local_address)

        remote_address = self.remote_address
        assert isinstance(remote_address, str)
        command = 'sudo nc -U "{}"'.format(remote_address)

        # reusing the same timeout value that is used by the parent class
        chan = self.transport.open_session(timeout=sshtunnel.TUNNEL_TIMEOUT)
        chan.exec_command(command)

        self.logger.log(sshtunnel.TRACE_LEVEL,
                        '{0} connected'.format(self.info))
        try:
            self._redirect(chan)
        except socket.error:
            # Sometimes a RST is sent and a socket error is raised, treat this
            # exception. It was seen that a 3way FIN is processed later on, so
            # no need to make an ordered close of the connection here or raise
            # the exception beyond this point...
            self.logger.log(sshtunnel.TRACE_LEVEL,
                            '{0} sending RST'.format(self.info))
        except Exception as e:
            self.logger.log(sshtunnel.TRACE_LEVEL,
                            '{0} error: {1}'.format(self.info, repr(e)))
        finally:
            chan.close()
            self.request.close()
            self.logger.log(sshtunnel.TRACE_LEVEL,
                            '{0} connection closed.'.format(self.info))

# pylint: enable=protected-access


class SSHTunnelForwarder(sshtunnel.SSHTunnelForwarder):

    daemon_forward_servers = True  #: flag tunnel threads in daemon mode
    daemon_transport = True  #: flag SSH transport thread in daemon mode

    def __init__(self, ssh_client, **kwargs):
        self.ssh_client = ssh_client
        params = self._merge_parameters(self._get_connect_parameters(),
                                        **kwargs)
        super(SSHTunnelForwarder, self).__init__(**params)

    def _merge_parameters(self, *dicts, **kwargs):
        result = {}
        for d in dicts + (kwargs,):
            if d:
                result.update((k, v) for k, v in d.items() if v is not None)
        return result

    @staticmethod
    def _consolidate_auth(ssh_password=None,
                          ssh_pkey=None,
                          ssh_pkey_password=None,
                          allow_agent=True,
                          host_pkey_directories=None,
                          logger=None):
        return None, None

    def _get_connect_parameters(self):
        parameters = self.ssh_client.setup_connect_parameters()
        return dict(ssh_address_or_host=parameters['hostname'],
                    ssh_username=parameters.get('username'),
                    ssh_password=parameters.get('password'),
                    ssh_pkey=parameters.get('pkey'),
                    ssh_port=parameters.get('port'),
                    ssh_private_key_password=parameters.get('passphrase'),
                    compression=parameters.get('compress'),
                    allow_agent=parameters.get('allow_agent'))

    def _connect_to_gateway(self):
        # pylint: disable=attribute-defined-outside-init
        self._transport = self._get_transport()

    def _get_transport(self):
        return self.ssh_client.connect().get_transport()

    def _stop_transport(self, force=True):
        if self.is_active:
            del self._transport
            assert not self.is_active
        super(SSHTunnelForwarder, self)._stop_transport(force=force)

    @staticmethod
    def _get_binds(bind_address, bind_addresses, is_remote=False):
        addr_kind = 'remote' if is_remote else 'local'

        if not bind_address and not bind_addresses:
            if is_remote:
                raise ValueError("No {0} bind addresses specified. Use "
                                 "'{0}_bind_address' or '{0}_bind_addresses'"
                                 " argument".format(addr_kind))
            else:
                return []
        elif bind_address and bind_addresses:
            raise ValueError("You can't use both '{0}_bind_address' and "
                             "'{0}_bind_addresses' arguments. Use one of "
                             "them.".format(addr_kind))
        if bind_address:
            bind_addresses = [bind_address]
        if not is_remote:
            # Add random port if missing in local bind
            for (i, local_bind) in enumerate(bind_addresses):
                if isinstance(local_bind, tuple) and len(local_bind) == 1:
                    bind_addresses[i] = (local_bind[0], 0)
        # check_addresses(bind_addresses, is_remote)
        return bind_addresses

    def _make_ssh_forward_handler_class(self, remote_address_):
        """
        Make SSH Handler class
        """
        if isinstance(remote_address_, tuple):
            return super(
                SSHTunnelForwarder, self)._make_ssh_forward_handler_class(
                    remote_address_)

        class Handler(SSHUnixForwardHandler):
            transport = self._transport
            remote_address = remote_address_
            logger = self.logger
        return Handler


def parse_url(url):
    if isinstance(url, urllib.parse.ParseResult):
        return url
    else:
        return urllib.parse.urlparse(url)


def binding_address(url):
    url = parse_url(url)
    if url.netloc:
        # Retains only scheme and netloc
        return (url.hostname, url.port)
    elif url.path:
        # Retains only scheme and path
        return url.path

    raise ValueError('Invalid URL: {!r}'.format(url))


def binding_url(address):
    if isinstance(address, tuple):
        try:
            hostname, = address
        except ValueError:
            hostname, port = address
        return 'tcp://{hostname}:{port}'.format(hostname=hostname,
                                                port=port)

    elif isinstance(address, str):
        return 'unix://{path}'.format(path=address)

    raise TypeError('Invalid address type: {!r}'.format(address))



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log
import netaddr

import tobiko
from tobiko.openstack.nova import _client
from tobiko.shell import ssh
from tobiko.shell import ping


LOG = log.getLogger(__name__)


def list_server_ip_addresses(server, network_name=None, ip_version=None,
                             address_type=None, check_connectivity=False,
                             ssh_client=None, count=None):
    ips = tobiko.Selection()
    for _network_name, addresses in server.addresses.items():
        if count and len(ips) == count:
            break

        # check network name
        if network_name and network_name != _network_name:
            continue

        for address in addresses:
            if check_server_ip_address(address,
                                       ip_version=ip_version,
                                       address_type=address_type):
                ips.append(netaddr.IPAddress(address['addr'],
                                             version=address['version']))

    # check ICMP connectivity
    if check_connectivity:
        ips = ping.list_reachable_hosts(ips, ssh_client=ssh_client)

    return ips


def check_server_ip_address(address, ip_version=None, address_type=None):
    if ip_version and ip_version != address['version']:
        return False

    # check IP address type
    if address_type:
        try:
            if address_type != address['OS-EXT-IPS:type']:
                return False
        except KeyError as ex:
            raise ValueError("Unable to get IP type from server address "
                             f"'{address}'") from ex

    return True


def find_server_ip_address(server, unique=False, **kwargs):
    count = unique and 2 or 1
    addresses = list_server_ip_addresses(server=server, count=count, **kwargs)
    return unique and addresses.unique or addresses.first


class HasServerMixin(_client.HasNovaClientMixin):

    @property
    def server_id(self):
        raise NotImplementedError

    @property
    def server(self):
        return _client.get_server(self.server_id)

    @property
    def server_name(self):
        return self.server.name

    @property
    def server_ips(self):
        return self.list_server_ips()

    def list_server_ips(self, **kwargs):
        return list_server_ip_addresses(server=self.server, **kwargs)

    @property
    def server_fixed_ips(self):
        return self.list_server_ips(address_type='fixed', check=True)

    @property
    def server_floating_ip(self):
        floating_ips = self.list_server_ips(address_type='floating', count=1)
        if floating_ips:
            return floating_ips.first
        else:
            return None

    @property
    def has_floating_ip(self):
        return bool(self.server_floating_ip)

    @property
    def server_console_output(self):
        return super(HasServerMixin, self).get_server_console_output(
            server=self.server_id)

    @property
    def server_public_ip(self):
        return self.server_floating_ip or self.server_fixed_ips.first

    @property
    def ssh_client(self):
        return ssh.ssh_client(host=self.server_public_ip,
                              username=self.ssh_username,
                              password=self.ssh_password,
                              proxy_client=self.ssh_proxy_client)

    ssh_proxy_client = None
    ssh_username = None
    ssh_password = None


def get_server_id(server: _client.ServerType) -> str:
    if isinstance(server, str):
        return server
    else:
        return server.id



"""A client for communicating with a Podman varlink service."""

from __future__ import absolute_import

import errno
import logging
import os
from urllib.parse import urlparse

from varlink import Client as VarlinkClient
from varlink import VarlinkError

from .libs import cached_property
from .libs.containers import Containers
from .libs.errors import error_factory
from .libs.images import Images
from .libs.system import System
from .libs.tunnel import Context, Portal, Tunnel
from .libs.pods import Pods


class BaseClient():
    """Context manager for API workers to access varlink."""

    def __init__(self, context):
        """Construct Client."""
        self._client = None
        self._iface = None
        self._context = context

    def __call__(self):
        """Support being called for old API."""
        return self

    @classmethod
    def factory(cls,
                uri=None,
                interface='io.podman',
                *_args,
                **kwargs):
        """Construct a Client based on input."""
        # pylint: disable=keyword-arg-before-vararg
        log_level = os.environ.get('LOG_LEVEL')
        if log_level is not None:
            logging.basicConfig(level=logging.getLevelName(log_level.upper()))

        if uri is None:
            raise ValueError('uri is required and cannot be None')
        if interface is None:
            raise ValueError('interface is required and cannot be None')

        unsupported = set(kwargs.keys()).difference(
            ('uri', 'interface', 'remote_uri', 'identity_file',
             'ignore_hosts', 'known_hosts'))
        if unsupported:
            raise ValueError('Unknown keyword arguments: {}'.format(
                ', '.join(unsupported)))

        local_path = urlparse(uri).path
        if local_path == '':
            raise ValueError('path is required for uri,'
                             ' expected format "unix://path_to_socket"')

        if kwargs.get('remote_uri') is None:
            return LocalClient(Context(uri, interface))

        required = ('{} is required, expected format'
                    ' "ssh://user@hostname[:port]/path_to_socket".')

        # Remote access requires the full tuple of information
        if kwargs.get('remote_uri') is None:
            raise ValueError(required.format('remote_uri'))

        remote = urlparse(kwargs['remote_uri'])
        if remote.username is None:
            raise ValueError(required.format('username'))
        if remote.path == '':
            raise ValueError(required.format('path'))
        if remote.hostname is None:
            raise ValueError(required.format('hostname'))

        return RemoteClient(
            Context(
                uri,
                interface,
                local_path,
                remote.path,
                remote.username,
                remote.hostname,
                remote.port,
                kwargs.get('identity_file'),
                kwargs.get('ignore_hosts'),
                kwargs.get('known_hosts'),
            ))


class LocalClient(BaseClient):
    """Context manager for API workers to access varlink."""

    def __enter__(self):
        """Enter context for LocalClient."""
        self._client = VarlinkClient(address=self._context.uri)
        self._iface = self._client.open(self._context.interface)
        return self._iface

    def __exit__(self, e_type, e, e_traceback):
        """Cleanup context for LocalClient."""
        if hasattr(self._client, 'close'):
            # pylint: disable=no-member
            self._client.close()
        self._iface.close()

        if isinstance(e, VarlinkError):
            raise error_factory(e)


class RemoteClient(BaseClient):
    """Context manager for API workers to access remote varlink."""

    def __init__(self, context):
        """Construct RemoteCLient."""
        super().__init__(context)
        self._portal = Portal()

    def __enter__(self):
        """Context manager for API workers to access varlink."""
        tunnel = self._portal.get(self._context.uri)
        if tunnel is None:
            tunnel = Tunnel(self._context).bore()
            self._portal[self._context.uri] = tunnel

        try:
            self._client = VarlinkClient(address=self._context.uri)
            self._iface = self._client.open(self._context.interface)
            return self._iface
        except Exception:
            tunnel.close()
            raise

    def __exit__(self, e_type, e, e_traceback):
        """Cleanup context for RemoteClient."""
        if hasattr(self._client, 'close'):
            # pylint: disable=no-member
            self._client.close()
        self._iface.close()

        # set timer to shutdown ssh tunnel
        # self._portal.get(self._context.uri).close()
        if isinstance(e, VarlinkError):
            raise error_factory(e)


class Client():
    """A client for communicating with a Podman varlink service.

    Example:

        >>> import podman
        >>> c = podman.Client()
        >>> c.system.versions

    Example remote podman:

        >>> import podman
        >>> c = podman.Client(uri='unix:/tmp/podman.sock',
                              remote_uri='ssh://user@host/run/podman/io.podman',
                              identity_file='~/.ssh/id_rsa')
    """

    def __init__(self,
                 uri='unix:/run/podman/io.podman',
                 interface='io.podman',
                 **kwargs):
        """Construct a podman varlink Client.

        uri from default systemd unit file.
        interface from io.podman.varlink, do not change unless
            you are a varlink guru.
        """
        self._client = BaseClient.factory(uri, interface, **kwargs)

        address = "{}-{}".format(uri, interface)
        # Quick validation of connection data provided
        try:
            if not System(self._client).ping():
                raise ConnectionRefusedError(
                    errno.ECONNREFUSED,
                    ('Failed varlink connection "{}"').format(address))
        except FileNotFoundError as ex:
            raise ConnectionError(
                errno.ECONNREFUSED,
                ('Failed varlink connection "{}".'
                 ' Is podman socket or service running?'
                 ).format(address)) from ex

    def __enter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""

    @cached_property
    def system(self):
        """Manage system model for podman."""
        return System(self._client)

    @cached_property
    def images(self):
        """Manage images model for libpod."""
        return Images(self._client)

    @cached_property
    def containers(self):
        """Manage containers model for libpod."""
        return Containers(self._client)

    @cached_property
    def pods(self):
        """Manage pods model for libpod."""
        return Pods(self._client)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.openstack import designate
from tobiko.openstack import heat
from tobiko.openstack.stacks import _hot


class DesignateZoneStackFixture(heat.HeatStackFixture):
    template = _hot.heat_template_file('designate/zone.yaml')

    @property
    def zone_name(self) -> str:
        return tobiko.get_fixture_name(self).lower() + '.'

    @property
    def zone_details(self) -> typing.Mapping[str, typing.Any]:
        return designate.get_designate_zone(self.zone_id)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import collections
import json
import re
import typing

import pytest
from oslo_log import log
import testtools

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import ip
from tobiko.openstack import neutron
from tobiko.openstack import nova
from tobiko.openstack import stacks
from tobiko.openstack import topology


LOG = log.getLogger(__name__)
CONF = config.CONF


@pytest.mark.minimal
class RouterTest(testtools.TestCase):
    """Test Neutron routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    def setUp(self):
        super(RouterTest, self).setUp()
        if not self.stack.network_stack.has_gateway:
            tobiko.skip_test(
                f"Stack {self.stack.network_stack.stack_name} has no gateway")

    @property
    def router(self):
        return self.stack.network_stack.gateway_details

    @property
    def ipv4_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv4:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv4 subnet")
        return self.stack.network_stack.ipv4_subnet_gateway_ip

    @property
    def ipv6_subnet_gateway_ip(self):
        if not self.stack.network_stack.has_ipv6:
            tobiko.skip_test(f"Stack {self.stack.network_stack.stack_name} "
                             "has no ipv6 subnet")
        return self.stack.network_stack.ipv6_subnet_gateway_ip

    @property
    def external_gateway_ips(self):
        return self.stack.network_stack.external_gateway_ips

    @property
    def router_ips(self):
        return tobiko.select([self.ipv4_subnet_gateway_ip,
                              self.ipv6_subnet_gateway_ip] +
                             self.external_gateway_ips)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv4_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv4_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    @tobiko.retry_test_case(interval=30.)
    def test_internal_router_ipv6_interface_is_reachable(self):
        ping.assert_reachable_hosts([self.ipv6_subnet_gateway_ip],
                                    ssh_client=self.stack.ssh_client)

    def test_ipv4_subnet_gateway_ip(self):
        self.assertEqual(4, self.ipv4_subnet_gateway_ip.version)
        self.assertIn(self.ipv4_subnet_gateway_ip,
                      self.stack.network_stack.ipv4_gateway_addresses)

    def test_ipv6_subnet_gateway_ip(self):
        self.assertEqual(6, self.ipv6_subnet_gateway_ip.version)
        self.assertIn(self.ipv6_subnet_gateway_ip,
                      self.stack.network_stack.ipv6_gateway_addresses)

    @pytest.mark.flaky(reruns=3, reruns_delay=60)
    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        router = self.router
        if router.get('distributed'):
            # TODO(fressi): check whenever a DVR router can be scheduled
            self.skipTest("Router is distributed")
        router_agent = neutron.find_l3_agent_hosting_router(router['id'],
                                                            unique=True)
        self._check_routers_namespace_on_host(router_agent['host'])

    def _check_routers_namespace_on_host(self, hostname, state="master"):
        namespace_ips = self._get_router_ips_from_namespaces(hostname)
        missing_ips = set(self.router_ips) - set(namespace_ips)
        if state == "master":
            self.assertFalse(missing_ips)
        else:
            self.assertTrue(missing_ips)

    def _get_router_ips_from_namespaces(self, hostname):
        agent_host = topology.get_openstack_node(hostname=hostname)
        router_namespaces = ["qrouter-%s" % self.router['id']]
        if self.router.get('distributed'):
            router_namespaces.append("snat-%s" % self.router['id'])
        host_namespaces = ip.list_network_namespaces(
            ssh_client=agent_host.ssh_client)
        ips = []
        for router_namespace in router_namespaces:
            self.assertIn(router_namespace, host_namespaces)
            ips += ip.list_ip_addresses(
                scope='global', network_namespace=router_namespace,
                ssh_client=agent_host.ssh_client)
        return ips


@neutron.skip_if_missing_networking_extensions('l3-ha')
@neutron.skip_if_missing_networking_agents(binary='neutron-l3-agent',
                                           count=2)
class L3HARouterTest(RouterTest):
    """Test Neutron HA routers"""

    #: Resources stack with Nova server to send messages to
    stack = tobiko.required_fixture(stacks.L3haServerStackFixture)

    @neutron.skip_if_missing_networking_extensions('l3_agent_scheduler')
    def test_router_is_scheduled_on_l3_agents(self):
        master_agent, backup_agents = (
            neutron.wait_for_master_and_backup_agents(self.router['id']))
        self.assertGreaterEqual(len(backup_agents), 1)
        self._check_routers_namespace_on_host(master_agent['host'])
        for backup_agent in backup_agents:
            self._check_routers_namespace_on_host(
                backup_agent['host'], state="backup")


class RouterNamespaceTestBase:

    server_stack = tobiko.required_fixture(stacks.CirrosServerStackFixture)

    host_groups = ['overcloud', 'compute', 'controller']

    @property
    def hostnames(self) -> typing.List[str]:
        return sorted(
            node.hostname
            for node in topology.list_openstack_nodes(group=self.host_groups))

    @property
    def router_stack(self) -> stacks.RouterStackFixture:
        return self.network_stack.gateway_stack

    @property
    def network_stack(self) -> stacks.NetworkStackFixture:
        return self.server_stack.network_stack

    @property
    def router_id(self) -> str:
        return self.router_stack.router_id

    @property
    def router_details(self) -> neutron.RouterType:
        return self.router_stack.router_details

    @property
    def router_namespace(self) -> str:
        return neutron.get_ovs_router_namespace(self.router_id)


@neutron.skip_unless_is_ovs()
class OvsRouterNamespaceTest(RouterNamespaceTestBase, testtools.TestCase):

    def test_router_namespace(self):
        """Check router namespace is being created on cloud hosts

        When A VM running in a compute node is expected to route
        packages to a router, a network namespace is expected to exist on
        compute name that is named after the router ID
        """
        ping.assert_reachable_hosts([self.server_stack.floating_ip_address])
        topology.assert_namespace_in_hosts(self.router_namespace,
                                           hostnames=self.hostnames)


@neutron.skip_if_missing_networking_extensions('dvr')
class DvrRouterStackFixture(stacks.RouterStackFixture):
    distributed = True


class DvrNetworkStackFixture(stacks.NetworkStackFixture):
    gateway_stack = tobiko.required_fixture(DvrRouterStackFixture,
                                            setup=False)


class DvrServerStackFixture(stacks.CirrosServerStackFixture):
    network_stack = tobiko.required_fixture(DvrNetworkStackFixture,
                                            setup=False)


L3_AGENT_MODE_DVR = re.compile(r'^dvr_')


def is_l3_agent_mode_dvr(agent_mode: str) -> bool:
    return L3_AGENT_MODE_DVR.match(agent_mode) is not None


@neutron.skip_if_missing_networking_extensions('dvr')
class DvrRouterNamespaceTest(RouterNamespaceTestBase, testtools.TestCase):

    server_stack = tobiko.required_fixture(DvrServerStackFixture, setup=False)

    def setUp(self):
        super().setUp()
        if self.legacy_hostnames:
            self.skipTest(f'Host(s) {self.legacy_hostnames!r} with legacy '
                          'L3 agent mode')

    host_groups = ['compute']

    @property
    def snat_namespace(self) -> str:
        return f'snat-{self.router_id}'

    _agent_modes: typing.Optional[typing.Dict[str, typing.List[str]]] = None

    @property
    def agent_modes(self) \
            -> typing.Dict[str, typing.List[str]]:
        if self._agent_modes is None:
            self._agent_modes = collections.defaultdict(list)
            for node in topology.list_openstack_nodes(
                    hostnames=self.hostnames):
                self._agent_modes[node.l3_agent_mode].append(node.name)
            agent_modes_dump = json.dumps(self._agent_modes,
                                          indent=4, sort_keys=True)
            LOG.info(f"Got L3 agent modes:\n{agent_modes_dump}")
        return self._agent_modes

    @property
    def legacy_hostnames(self) -> typing.List[str]:
        return self.agent_modes['legacy']

    @property
    def dvr_hostnames(self) -> typing.List[str]:
        return self.agent_modes['dvr'] + self.agent_modes['dvr_no_external']

    @property
    def dvr_snat_hostnames(self) -> typing.List[str]:
        return self.agent_modes['dvr_snat']

    @config.skip_if_prevent_create()
    def test_1_dvr_router_without_server(self):
        if self.dvr_hostnames:
            self.cleanup_stacks()
            self.setup_router()
            topology.assert_namespace_not_in_hosts(
                self.router_namespace,
                self.snat_namespace,
                hostnames=self.dvr_hostnames)
        else:
            self.skipTest(f'All hosts {self.hostnames!r} have '
                          'dvr_snat L3 agent mode')

    def test_2_dvr_snat_router_namespaces(self):
        if self.dvr_snat_hostnames:
            self.setup_router()
            topology.wait_for_namespace_in_hosts(
                self.router_namespace,
                self.snat_namespace,
                hostnames=self.dvr_snat_hostnames)
        else:
            self.skipTest(f'Any host {self.hostnames!r} has '
                          'dvr_snat L3 agent mode')

    def test_3_dvr_router_namespace_with_server(self):
        self.setup_server()
        self.wait_for_namespace_in_hypervisor_host()

    def wait_for_namespace_in_hypervisor_host(self):
        hypervisor_hostname = self.server_stack.hypervisor_hostname
        agent_mode = topology.get_l3_agent_mode(hostname=hypervisor_hostname)
        LOG.info(f"Hypervisor host '{hypervisor_hostname}' has DVR agent "
                 f"mode: '{agent_mode}'")
        topology.wait_for_namespace_in_hosts(self.router_namespace,
                                             hostnames=[hypervisor_hostname])

    def test_4_server_is_reachable(self):
        self.setup_server()
        try:
            self.server_stack.assert_is_reachable()
        except ping.PingFailed:
            server_id = self.server_stack.server_id
            server_log = nova.get_console_output(server_id=server_id)
            LOG.exception(f"Unable to reach server {server_id}...\n"
                          f"{server_log}\n")
            self.wait_for_namespace_in_hypervisor_host()
            nova.reboot_server(server=server_id)
            self.server_stack.assert_is_reachable()

    def cleanup_stacks(self):
        for stack in [self.server_stack,
                      self.network_stack,
                      self.router_stack]:
            tobiko.cleanup_fixture(stack)

    def setup_router(self):
        router = tobiko.setup_fixture(self.router_stack).router_details
        router_dump = json.dumps(router, indent=4, sort_keys=True)
        LOG.debug(f"Testing DVR router namespace: {router['id']}:\n"
                  f"{router_dump}\n")
        self.assertTrue(router['distributed'])

    def setup_network(self):
        self.setup_router()
        tobiko.setup_fixture(self.network_stack)

    def setup_server(self):
        self.setup_network()
        tobiko.setup_fixture(self.server_stack)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from tobiko.podman import _client
from tobiko.podman import _shell
from tobiko.podman import _exception

Podman_Version_3 = _client.podman_version_3
PodmanClientFixture = _client.PodmanClientFixture
get_podman_client = _client.get_podman_client
list_podman_containers = _client.list_podman_containers
podman_client = _client.podman_client

discover_podman_socket = _shell.discover_podman_socket
is_podman_running = _shell.is_podman_running

PodmanError = _exception.PodmanError
PodmanSocketNotFoundError = _exception.PodmanSocketNotFoundError



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import re

from oslo_log import log
import netaddr

import tobiko


LOG = log.getLogger(__name__)


def parse_ping_statistics(output, begin_interval=None, end_interval=None):
    lines = output.split('\n')
    line_it = iter(lines)
    try:
        source, destination = parse_ping_header(line_it)
    except Exception as ex:
        LOG.debug('Error parsing ping output header: %s', ex)
        source = destination = None

    try:
        transmitted, received, errors = parse_ping_footer(line_it)
    except Exception as ex:
        LOG.debug('Error parsing ping output footer: %s', ex)
        transmitted = received = errors = 0

    return PingStatistics(source=source, destination=destination,
                          transmitted=transmitted, received=received,
                          undelivered=errors, end_interval=end_interval,
                          begin_interval=begin_interval)


def parse_ping_header(line_it):
    for line in line_it:
        if line.startswith('PING '):
            header = line
            break
    else:
        raise ValueError('Ping output header not found')

    header_fields = [f.strip() for f in header.split()]
    header_fields = [(f[:-1] if f.endswith(':') else f)
                     for f in header_fields]
    destination = header_fields[2]
    while destination and destination[0] == '(':
        destination = destination[1:]
    while destination and destination[-1] == ')':
        destination = destination[:-1]
    destination = netaddr.IPAddress(destination)

    field_it = iter(header_fields[3:])
    source = None

    for field in field_it:
        if field == 'from':
            source = next(field_it).strip()
            if source.endswith(':'):
                source = source[:-1]
            source = netaddr.IPAddress(source)
            break

    return source, destination


def parse_ping_footer(line_it):
    for line in line_it:
        if line.startswith('--- ') and line.endswith(' ping statistics ---'):
            # parse footer
            footer = next(line_it).strip()
            break
    else:
        raise ValueError('Ping output footer not found')

    transmitted = received = errors = 0
    for field in footer.split(','):
        try:
            if 'transmitted' in field:
                transmitted = extract_integer(field)
            elif 'received' in field:
                received = extract_integer(field)
            elif 'error' in field:
                errors = extract_integer(field)
        except Exception as ex:
            LOG.exception('Error parsing ping output footer: %s', ex)
    return transmitted, received, errors


def extract_integer(field):
    for number in extract_integers(field):
        return number
    raise ValueError("Integer not found in {!r}".format(field))


MATCH_NUMBERS_RE = re.compile('([0-9]).')


def extract_integers(field):
    for match_obj in MATCH_NUMBERS_RE.finditer(field):
        yield int(field[match_obj.start():match_obj.end()])


class PingStatistics(object):
    """Ping command statistics

    """

    def __init__(self,
                 source=None,
                 destination=None,
                 transmitted: int = 0,
                 received: int = 0,
                 undelivered: int = 0,
                 begin_interval=None,
                 end_interval=None):
        self.source = source
        self.destination = destination
        self.transmitted = transmitted
        self.received = received
        self.undelivered = undelivered
        self.begin_interval = begin_interval
        self.end_interval = end_interval

    @property
    def unreceived(self) -> int:
        return max(0, self.transmitted - self.received)

    @property
    def delivered(self) -> int:
        return max(0, self.transmitted - self.undelivered)

    @property
    def loss(self) -> float:
        if self.transmitted > 0:
            return float(self.unreceived) / float(self.transmitted)
        else:
            return 0.

    def __bool__(self):
        return bool(self.received)

    def __add__(self, other):
        begin_interval = min(i for i in [self.begin_interval,
                                         other.begin_interval] if i)
        end_interval = max(i for i in [self.end_interval,
                                       other.end_interval] if i)
        return PingStatistics(
            source=self.source or other.source,
            destination=self.destination or other.destination,
            transmitted=self.transmitted + other.transmitted,
            received=self.received + other.received,
            undelivered=self.undelivered + other.undelivered,
            begin_interval=begin_interval,
            end_interval=end_interval)

    def __repr__(self):
        return "PingStatistics({!s})".format(
            ", ".join("{!s}={!r}".format(k, v)
                      for k, v in self.__dict__.items()))

    def assert_transmitted(self):
        if not self.transmitted:
            tobiko.fail(f"{self.transmitted} package(s) has been transmitted "
                        f"to {self.destination}")

    def assert_not_transmitted(self):
        if self.transmitted:
            tobiko.fail(f"{self.transmitted} package(s) has been "
                        f"transmitted to {self.destination}")

    def assert_replied(self):
        if not self.received:
            tobiko.fail(f"{self.received} reply package(s) has been "
                        f"received from {self.destination}")

    def assert_not_replied(self):
        if self.received:
            tobiko.fail(f"{self.received} reply package(s) has been received "
                        f"from {self.destination}")



# Copyright (c) 2020 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import stacks
from tobiko.tests.functional.openstack.stacks import test_centos


@keystone.skip_unless_has_keystone_credentials()
class RedHatServerStackTest(test_centos.CentosServerStackTest):
    """Test Red Hat server instance"""

    #: Stack of resources with a server attached to a floating IP
    stack = tobiko.required_fixture(stacks.RedHatServerStackFixture)



# Copyright (c) 2022 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import functools
import os
import typing
import unittest

import testtools

from tobiko import run


def nested_test_case(test_method: typing.Callable[[testtools.TestCase], None]):

    @functools.wraps(test_method)
    def wrapper(self: unittest.TestCase):
        nested_counter = int(os.environ.get('NESTED_TEST_CASE', 0))
        if not nested_counter:
            os.environ['NESTED_TEST_CASE'] = str(nested_counter + 1)
            try:
                test_method(self)
            finally:
                if nested_counter:
                    os.environ['NESTED_TEST_CASE'] = str(nested_counter)
                else:
                    os.environ.pop('NESTED_TEST_CASE')
    return wrapper


class RunTestsTest(unittest.TestCase):

    @nested_test_case
    def test_run_tests(self):
        result = run.run_tests(__file__)
        self.assertGreater(result.testsRun, 0)

    @nested_test_case
    def test_run_tests_with_dir(self):
        test_dir = os.path.dirname(__file__)
        result = run.run_tests(test_path=test_dir)
        self.assertGreater(result.testsRun, 0)



# Copyright (c) 2019 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from keystoneauth1 import session as keystonesession
import mock

from tobiko.openstack import keystone
from tobiko.tests.unit import openstack


CREDENTIALS = keystone.keystone_credentials(
    api_version=3,
    auth_url='http://127.0.0.1:4000/identiy/v3',
    username='demo',
    project_name='demo',
    password='this is a secret',
    user_domain_name='Default',
    project_domain_name='Default')


DEFAULT_CREDENTIALS = keystone.keystone_credentials(
    api_version=2,
    auth_url='http://127.0.0.1:5000/identiy/v2.0',
    username='admin',
    project_name='admin',
    password='this is a secret')


class CredentialsFixture(keystone.KeystoneCredentialsFixture):

    def _get_credentials(self) -> keystone.KeystoneCredentials:
        return CREDENTIALS


class DefaultCredentialsFixture(CredentialsFixture):

    def _get_credentials(self) -> keystone.KeystoneCredentials:
        return DEFAULT_CREDENTIALS


class KeystoneSessionFixtureTest(openstack.OpenstackTest):

    def test_init(self,
                  credentials: keystone.KeystoneCredentialsType = None):
        # pylint: disable=protected-access
        session = keystone.KeystoneSessionFixture(credentials=credentials)
        self.assertIs(credentials, session._credentials)

    def test_init_with_credentials(self):
        self.test_init(credentials=CREDENTIALS)

    def test_init_with_credentials_fixture(self):
        self.test_init(credentials=CredentialsFixture())

    def test_init_with_credentials_fixture_type(self):
        self.test_init(credentials=CredentialsFixture)

    def test_setup(self,
                   credentials: keystone.KeystoneCredentialsType = None):
        session = keystone.KeystoneSessionFixture(credentials=credentials)
        session.setUp()
        self.assertEqual(keystone.keystone_credentials(credentials),
                         session.credentials)

    def test_setup_with_credentials(self):
        self.test_setup(credentials=CREDENTIALS)

    def test_setup_with_credentials_fixture(self):
        self.test_setup(credentials=CredentialsFixture())

    def test_setup_with_credentials_fixture_type(self):
        self.test_setup(credentials=CredentialsFixture)


class KeystoneSessionManagerTest(openstack.OpenstackTest):

    def test_init(self):
        manager = keystone.KeystoneSessionManager()
        self.assertTrue(manager)
        self.assertEqual({}, manager.sessions)

    def test_get_session(self,
                         credentials: keystone.KeystoneCredentialsType = None,
                         shared=True):
        manager = keystone.KeystoneSessionManager()

        session = manager.get_session(credentials=credentials,
                                      shared=shared)
        self.assertIsNotNone(session.credentials)
        self.assertIs(keystone.keystone_credentials(credentials),
                      session.credentials)
        self.assertIsInstance(session, keystone.KeystoneSessionFixture)
        shared_session = manager.get_session(credentials=credentials)
        if shared in [True, None]:
            self.assertIs(session, shared_session)
        else:
            self.assertIsNot(session, shared_session)

    def test_get_session_with_credentials(self):
        self.test_get_session(credentials=CREDENTIALS)

    def test_get_session_with_not_shared(self):
        self.test_get_session(shared=False)

    def test_get_session_with_credentials_fixture(self):
        self.test_get_session(credentials=CredentialsFixture())

    def test_get_session_with_credentials_fixture_type(self):
        self.test_get_session(credentials=CredentialsFixture)

    def test_get_session_with_init_session(self):
        mock_session = mock.MagicMock(spec=keystone.KeystoneSessionFixture)
        self.assertIsInstance(mock_session, keystone.KeystoneSessionFixture)
        init_session = mock.MagicMock(return_value=mock_session)
        manager = keystone.KeystoneSessionManager()
        session = manager.get_session(credentials=CREDENTIALS,
                                      init_session=init_session)
        self.assertIs(mock_session, session)
        init_session.assert_called_once_with(CREDENTIALS)


class GetKeystomeSessionTest(openstack.OpenstackTest):

    def test_get_keystone_session(
            self,
            credentials: keystone.KeystoneCredentialsType = None,
            shared=True):
        session1 = keystone.get_keystone_session(
            credentials=credentials,
            shared=shared)
        session2 = keystone.get_keystone_session(credentials=credentials,
                                                 shared=shared)
        if shared:
            self.assertIs(session1, session2)
        else:
            self.assertIsNot(session1, session2)
        self.assertIsInstance(session1, keystonesession.Session)
        self.assertIsInstance(session2, keystonesession.Session)

    def test_get_keystone_session_with_not_shared(self):
        self.test_get_keystone_session(shared=False)

    def test_get_keystone_session_with_credentials(self):
        credentials = keystone.default_keystone_credentials()
        self.test_get_keystone_session(credentials=credentials)



# Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import asyncio
import typing

import tobiko
from tobiko.actors import _actor


A = typing.TypeVar('A', bound=_actor.ActorBase)


class ActorManager(tobiko.SharedFixture, tobiko.FixtureManager):
    pass


def actor_manager(obj: ActorManager = None) -> ActorManager:
    if obj is None:
        return tobiko.get_fixture(ActorManager)
    return tobiko.check_valid_type(obj, ActorManager)


ActorType = typing.Union[A, typing.Type[A]]


def start_actor(obj: ActorType,
                fixture_id: typing.Optional[str] = None,
                manager: ActorManager = None) -> _actor.ActorRef[A]:
    return tobiko.setup_fixture(obj,
                                fixture_id=fixture_id,
                                manager=manager).ref


async def setup_actor(obj: ActorType,
                      fixture_id: typing.Optional[str] = None,
                      manager=None,
                      timeout: tobiko.Seconds = None) -> _actor.ActorRef[A]:
    actor = tobiko.setup_fixture(obj,
                                 fixture_id=fixture_id,
                                 manager=manager)
    await asyncio.wait_for(actor.setup_actor_future,
                           timeout=timeout)
    return actor.ref


async def stop_actor(obj: ActorType,
                     fixture_id: typing.Optional[str] = None,
                     manager=None) -> _actor.ActorRef[A]:
    return tobiko.cleanup_fixture(obj,
                                  fixture_id=fixture_id,
                                  manager=manager).ref


async def cleanup_actor(obj: ActorType,
                        fixture_id: typing.Optional[str] = None,
                        manager=None,
                        timeout: tobiko.Seconds = None) -> _actor.ActorRef[A]:
    actor = tobiko.cleanup_fixture(obj,
                                   fixture_id=fixture_id,
                                   manager=manager)
    await asyncio.wait_for(actor.cleanup_actor_future,
                           timeout=timeout)
    return actor.ref



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import inspect
import os
import sys
import typing
import unittest

from oslo_log import log

import tobiko
from tobiko.run import _config
from tobiko.run import _find
from tobiko.run import _worker


LOG = log.getLogger(__name__)


def find_test_ids(test_path: typing.Union[str, typing.Iterable[str]],
                  test_filename: str = None,
                  python_path: typing.Iterable[str] = None,
                  forked: bool = None,
                  config: _config.RunConfigFixture = None) \
        -> typing.List[str]:
    config = _config.run_confing(config)
    test_files = _find.find_test_files(test_path=test_path,
                                       test_filename=test_filename,
                                       config=config)
    if not python_path:
        python_path = config.python_path
    if forked is None:
        forked = bool(config.forked)

    if forked:
        return forked_discover_test_ids(test_files=test_files,
                                        python_path=python_path)
    else:
        return discover_test_ids(test_files=test_files,
                                 python_path=python_path)


def discover_test_ids(test_files: typing.Iterable[str],
                      python_path: typing.Iterable[str] = None) \
        -> typing.List[str]:
    if not python_path:
        python_path = sys.path
    python_dirs = [os.path.realpath(p) + '/'
                   for p in python_path
                   if os.path.isdir(p)]
    test_ids: typing.List[str] = []
    for test_file in test_files:
        test_ids.extend(discover_file_test_ids(test_file=test_file,
                                               python_dirs=python_dirs))
    return test_ids


def discover_file_test_ids(test_file: str,
                           python_dirs: typing.Iterable[str]) \
        -> typing.List[str]:
    test_file = os.path.realpath(test_file)
    if not os.path.isfile(test_file):
        raise ValueError(f"Test file doesn't exist: '{test_file}'")

    if not test_file.endswith('.py'):
        raise ValueError(f"Test file hasn't .py suffix: '{test_file}'")

    for python_dir in python_dirs:
        if test_file.startswith(python_dir):
            module_name = test_file[len(python_dir):-3].replace('/', '.')
            return discover_module_test_ids(module_name)

    raise ValueError(f"Test file not in Python path: '{test_file}'")


def discover_module_test_ids(module_name: str) -> typing.List[str]:
    LOG.debug(f"Load test module '{module_name}'...")
    module = tobiko.load_module(module_name)
    test_file = module.__file__
    LOG.debug("Inspect test module:\n"
              f"  module: '{module_name}'\n"
              f"  filename: '{test_file}'\n")
    test_ids: typing.List[str] = []
    for obj_name in dir(module):
        try:
            obj = getattr(module, obj_name)
        except AttributeError:
            LOG.warning("Error getting object "
                        f"'{module_name}.{obj_name}'",
                        exc_info=1)
            continue
        if (inspect.isclass(obj) and
                issubclass(obj, unittest.TestCase) and
                not inspect.isabstract(obj)):
            LOG.debug("Inspect test class members...\n"
                      f"  file: '{test_file}'\n"
                      f"  module: '{module_name}'\n"
                      f"  object: '{obj_name}'\n")
            for member_name in dir(obj):
                if member_name.startswith('test_'):
                    member_id = f"{module_name}.{obj_name}.{member_name}"
                    try:
                        member = getattr(obj, member_name)
                    except Exception:
                        LOG.error(f'Error getting "{member_id}"', exc_info=1)
                        continue
                    if not callable(member):
                        LOG.error("Class member is not callable: "
                                  f"'{member_id}'")
                        continue
                    test_ids.append(member_id)
    return test_ids


def forked_discover_test_ids(test_files: typing.Iterable[str],
                             python_path: typing.Iterable[str] = None) \
        -> typing.List[str]:
    results = [_worker.call_async(discover_test_ids,
                                  test_files=[test_file],
                                  python_path=python_path)
               for test_file in test_files]
    test_ids: typing.List[str] = []
    for result in results:
        test_ids.extend(result.get())
    return test_ids


def main(test_path: typing.Iterable[str] = None,
         test_filename: str = None,
         forked: bool = None,
         python_path: typing.Iterable[str] = None):
    if test_path is None:
        test_path = sys.argv[1:]
    try:
        test_ids = find_test_ids(test_path=test_path,
                                 test_filename=test_filename,
                                 forked=forked,
                                 python_path=python_path)
    except Exception as ex:
        sys.stderr.write(f'{ex}\n')
        sys.exit(1)
    else:
        output = ''.join(f'{test_id}\n'
                         for test_id in test_ids)
        sys.stdout.write(output)
        sys.exit(0)


if __name__ == '__main__':
    main()



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import json

from oslo_log import log

import tobiko
from tobiko.openstack.neutron import _agent
from tobiko.openstack.neutron import _client
from tobiko.openstack.neutron import _network
from tobiko.openstack.neutron import _port
from tobiko.openstack.neutron import _subnet


LOG = log.getLogger(__name__)


RouterType = typing.Dict[str, typing.Any]
RouterIdType = typing.Union[str, RouterType]


def get_router_id(router: RouterIdType) -> str:
    if isinstance(router, str):
        return router
    else:
        return router['id']


def get_router(router: RouterIdType, client=None, **params) -> RouterType:
    router_id = get_router_id(router)
    try:
        return _client.neutron_client(client).show_router(
            router_id, **params)['router']
    except _client.NotFound as ex:
        raise NoSuchRouter(id=router_id) from ex


def create_router(client: _client.NeutronClientType = None,
                  network: _network.NetworkIdType = None,
                  add_cleanup=True,
                  **params) -> RouterType:
    if 'external_gateway_info' not in params:
        if network is None:
            from tobiko.openstack import stacks
            network_id = stacks.get_floating_network_id()
        else:
            network_id = _network.get_network_id(network)
        params['external_gateway_info'] = dict(network_id=network_id)
    router = _client.neutron_client(client).create_router(
        body={'router': params})['router']
    if add_cleanup:
        tobiko.add_cleanup(cleanup_router, router=router, client=client)
    return router


def cleanup_router(router: RouterIdType,
                   client: _client.NeutronClientType = None):
    try:
        delete_router(router=router, client=client)
    except NoSuchRouter:
        pass


def delete_router(router: RouterIdType,
                  client: _client.NeutronClientType = None):
    router_id = get_router_id(router)
    try:
        _client.neutron_client(client).delete_router(router_id)
    except _client.NotFound as ex:
        raise NoSuchRouter(id=router_id) from ex


def wait_for_master_and_backup_agents(
        router: RouterIdType,
        unique_master: bool = True,
        timeout: tobiko.Seconds = None,
        interval: tobiko.Seconds = None) -> \
        typing.Tuple[typing.Dict, typing.List[typing.Dict]]:
    router_id = get_router_id(router)
    for attempt in tobiko.retry(timeout=timeout,
                                interval=interval,
                                default_timeout=300.,
                                default_interval=5.):
        router_agents = _agent.list_l3_agent_hosting_routers(router_id)
        master_agents = router_agents.with_items(ha_state='active')
        if master_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(master_agents)} master "
                "agent(s):\n"
                f"{json.dumps(master_agents, indent=4, sort_keys=True)}")
        backup_agents = router_agents.with_items(ha_state='standby')
        if backup_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(backup_agents)} backup "
                "agent(s)):\n"
                f"{json.dumps(backup_agents, indent=4, sort_keys=True)}")
        other_agents = [agent
                        for agent in router_agents
                        if (agent not in master_agents + backup_agents)]
        if other_agents:
            LOG.debug(
                f"Router '{router_id}' has {len(other_agents)} other "
                "agent(s):\n"
                f"{json.dumps(master_agents, indent=4, sort_keys=True)}")
        try:
            if unique_master:
                master_agent = master_agents.unique
            else:
                master_agent = master_agents.first
        except (tobiko.MultipleObjectsFound, tobiko.ObjectNotFound):
            attempt.check_limits()
        else:
            break
    else:
        raise RuntimeError("tobiko retry loop ended before break?")

    return master_agent, backup_agents


RouterInterfaceType = typing.Dict[str, typing.Any]


def add_router_interface(router: RouterIdType,
                         subnet: _subnet.SubnetIdType = None,
                         port: _port.PortIdType = None,
                         client: _client.NeutronClientType = None,
                         add_cleanup=True,
                         **params) -> RouterInterfaceType:
    router_id = get_router_id(router)
    if 'port_id' not in params and port is not None:
        params['port_id'] = _port.get_port_id(port)
    elif 'subnet_id' not in params and subnet is not None:
        params['subnet_id'] = _subnet.get_subnet_id(subnet)
    interface = _client.neutron_client(client).add_interface_router(
        router=router_id, body=params)
    if add_cleanup:
        tobiko.add_cleanup(cleanup_router_interface, router=router,
                           subnet=subnet, port=port, client=client)
    return interface


def cleanup_router_interface(router: RouterIdType,
                             subnet: _subnet.SubnetIdType = None,
                             port: _port.PortIdType = None,
                             client: _client.NeutronClientType = None,
                             **params):
    try:
        remove_router_interface(router=router, subnet=subnet,
                                port=port, client=client, **params)
    except tobiko.ObjectNotFound:
        pass


def remove_router_interface(router: RouterIdType,
                            subnet: _subnet.SubnetIdType = None,
                            port: _port.PortIdType = None,
                            client: _client.NeutronClientType = None,
                            **params):
    router_id = get_router_id(router)
    if 'port_id' not in params and port is not None:
        params['port_id'] = _port.get_port_id(port)
    elif 'subnet_id' not in params and subnet is not None:
        params['subnet_id'] = _subnet.get_subnet_id(subnet)
    try:
        _client.neutron_client(client).remove_interface_router(
            router=router_id, body=params)
    except _client.NotFound as ex:
        raise tobiko.ObjectNotFound() from ex


class NoSuchRouter(tobiko.ObjectNotFound):
    message = "No such router found for {id!r}"


def get_ovs_router_namespace(router: RouterIdType):
    return f"qrouter-{get_router_id(router)}"



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import tobiko
from tobiko.tests import unit
from tobiko import actors


class Greeter(actors.Actor):

    setup_called = False
    cleanup_called = False

    async def setup_actor(self):
        self.setup_called = True

    async def cleanup_actor(self):
        self.cleanup_called = True

    @actors.actor_method
    async def greet(self, whom: str, greeted: 'Greeted'):
        assert self.setup_called
        assert not self.cleanup_called
        if not whom:
            raise ValueError("'whom' parameter can't be empty")

        self.log.info(f"Hello {whom}!")
        await greeted.greeted(whom=whom, greeter=self)


class Greeted:

    greeter: typing.Optional[Greeter] = None
    whom: typing.Optional[str] = None

    async def greeted(self, whom: str, greeter: Greeter):
        self.greeter = greeter
        self.whom = whom


class ActorTest(unit.TobikoUnitTest):

    actor = tobiko.required_fixture(Greeter, setup=False)

    async def test_setup_actor(self):
        self.assertFalse(self.actor.setup_called)
        self.assertFalse(self.actor.cleanup_called)
        await actors.setup_actor(self.actor)
        self.assertTrue(self.actor.setup_called)
        self.assertFalse(self.actor.cleanup_called)

    async def test_cleanup_actor(self):
        self.assertFalse(self.actor.setup_called)
        self.assertFalse(self.actor.cleanup_called)
        await actors.cleanup_actor(self.actor)
        self.assertFalse(self.actor.setup_called)
        self.assertFalse(self.actor.cleanup_called)

    async def test_cleanup_actor_after_setup(self):
        await actors.setup_actor(self.actor)
        self.assertTrue(self.actor.setup_called)
        self.assertFalse(self.actor.cleanup_called)
        await actors.cleanup_actor(self.actor)
        self.assertTrue(self.actor.setup_called)
        self.assertTrue(self.actor.cleanup_called)

    async def test_ping_actor(self):
        ref = actors.start_actor(self.actor)
        result = await ref.ping_actor(self.id())
        self.assertEqual(self.id(), result)

    async def test_async_request(self):
        greeter = actors.start_actor(self.actor)
        self.assertIsInstance(greeter, actors.ActorRef[Greeter])
        greeted = Greeted()
        await greeter.greet(whom=self.id(), greeted=greeted)
        self.assertEqual(self.id(), greeted.whom)
        self.assertIs(self.actor, greeted.greeter)

    async def test_async_request_failure(self):
        greeter = actors.start_actor(self.actor)
        self.assertIsInstance(greeter, actors.ActorRef[Greeter])
        greeted = Greeted()

        try:
            await greeter.greet(whom="", greeted=greeted)
        except ValueError as ex:
            self.assertEqual("'whom' parameter can't be empty", str(ex))
        else:
            self.fail("Exception not raised")
        self.assertIsNone(greeted.whom)
        self.assertIsNone(greeted.greeter)



# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import itertools

from oslo_config import cfg

GROUP_NAME = "ping"
OPTIONS = [
    cfg.IntOpt('count',
               default=1,
               help="Number of ICMP messages to wait before ending "
                    "ping command execution"),
    cfg.IntOpt('deadline',
               default=5,
               help="Max seconds waited from ping command before "
                    "self terminating himself"),
    cfg.StrOpt('fragmentation',
               default=None,
               help="If False it will not allow ICMP messages to "
                    "be delivered in smaller fragments"),
    cfg.StrOpt('interval',
               default=1,
               help="Seconds of time interval between "
                    "consecutive before ICMP messages"),
    cfg.IntOpt('packet_size',
               default=None,
               help="Size in bytes of ICMP messages (including "
                    "headers and payload)"),
    cfg.IntOpt('timeout',
               default=300.,
               help="Maximum time in seconds a sequence of ICMP "
                    "messages is sent to a destination host before "
                    "reporting as a failure")]


def register_tobiko_options(conf):
    conf.register_opts(group=cfg.OptGroup('ping'), opts=OPTIONS)


def list_options():
    return [(GROUP_NAME, itertools.chain(OPTIONS))]



# Copyright (c) 2021 Red Hat
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack import octavia
from tobiko.openstack import stacks
from tobiko.shell import ssh
from tobiko.shell import sh
from tobiko import tripleo


LOG = log.getLogger(__name__)


@keystone.skip_if_missing_service(name='octavia')
@tripleo.skip_if_missing_overcloud
class OctaviaBasicFaultTest(testtools.TestCase):
    """Octavia fault scenario test.

    Create a load balancer with 2 members that run a server application,
    Create a client that is connected to the load balancer VIP port,
    Generate network traffic from the client to the load balanacer VIP.
    Restart the amphora's compute node to create a failover.
    Reach the members to make sure they are ready to be checked.
    Generate network traffic again to verify Octavia functionality.
    """
    loadbalancer_stack = tobiko.required_fixture(
        stacks.AmphoraIPv4LoadBalancerStack)

    listener_stack = tobiko.required_fixture(
        stacks.HttpRoundRobinAmphoraIpv4Listener)

    def setUp(self):
        # pylint: disable=no-member
        super(OctaviaBasicFaultTest, self).setUp()

        # Wait for Octavia objects to be active
        LOG.info('Waiting for member '
                 f'{self.listener_stack.server_stack.stack_name} and '
                 f'for member '
                 f'{self.listener_stack.other_server_stack.stack_name} '
                 f'to be created...')
        self.listener_stack.wait_for_active_members()

        self.loadbalancer_stack.wait_for_octavia_service()

        self.listener_stack.wait_for_members_to_be_reachable()

        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being provisioned
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except (octavia.RoundRobinException,
                    octavia.TrafficTimeoutError,
                    sh.ShellCommandFailed):
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise

    @property
    def amphora_ssh_client(self) -> ssh.SSHClientType:
        return self.listener_stack.amphora_ssh_client

    def test_reboot_amphora_compute_node(self):
        amphora_compute_host = octavia.get_amphora_compute_node(
            load_balancer=self.loadbalancer_stack.loadbalancer_id,
            port=self.listener_stack.lb_port,
            protocol=self.listener_stack.lb_protocol,
            ip_address=self.loadbalancer_stack.floating_ip_address)

        LOG.debug('Rebooting compute node...')

        # Reboot Amphora's compute node will initiate a failover
        amphora_compute_host.reboot_overcloud_node()

        LOG.debug('Compute node has been rebooted')

        # Wait for the LB to be updated
        try:
            self.loadbalancer_stack.wait_for_update_loadbalancer(timeout=30)

        except tobiko.RetryTimeLimitError:
            LOG.info('The restarted servers reached ACTIVE status after the'
                     ' LB finished its update process, hence no exception is'
                     ' being raised even though the update timeout was'
                     ' reached.')

        self.loadbalancer_stack.wait_for_active_loadbalancer()

        LOG.debug(f'Load Balancer {self.loadbalancer_stack.loadbalancer_id} is'
                  f' ACTIVE')

        # Wait for Octavia objects' provisioning status to be ACTIVE
        self.listener_stack.wait_for_active_members()

        # Verify Octavia functionality
        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being provisioned/migrated
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except (octavia.RoundRobinException,
                    octavia.TrafficTimeoutError,
                    sh.ShellCommandFailed):
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise

    def test_kill_amphora_agent(self):
        """Kill the MASTER amphora agent

        This test kills the amphora agent on the MASTER amphora.
        Killing the amphora agent will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        # Finding the amphora agent pid and kill it
        amp_agent_pid_command = (
            "ps -ef | awk '/amphora/{print $2}' | head -n 1")
        amp_agent_pid = octavia.run_command_on_amphora(
            command=amp_agent_pid_command,
            lb_id=self.loadbalancer_stack.loadbalancer_id,
            lb_fip=self.loadbalancer_stack.floating_ip_address)
        LOG.info(f'The amp_agent_pid is {amp_agent_pid}')

        octavia.run_command_on_amphora(
            command=f'kill -9 {amp_agent_pid}',
            lb_id=self.loadbalancer_stack.loadbalancer_id,
            lb_fip=self.loadbalancer_stack.floating_ip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def test_stop_keepalived(self):
        """Stop keepalived on MASTER amphora

        This test stops keepalived on the MASTER amphora.
        Stopping keepalived on the amphora will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        stop_keepalived_cmd = 'systemctl stop octavia-keepalived'

        octavia.run_command_on_amphora(
            command=stop_keepalived_cmd,
            lb_id=self.loadbalancer_stack.loadbalancer_id,
            lb_fip=self.loadbalancer_stack.floating_ip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def test_stop_haproxy(self):
        """Stop haproxy on MASTER amphora

        This test stops haproxy on the MASTER amphora.
        Stopping haproxy on the amphora will cause a failover.

        Octavia's functionality will be verified afterwards.
        """

        self._skip_if_not_active_standby()

        # Finding the amphora haproxy unit name and stop it
        amp_haproxy_unit_command = (
            "systemctl list-units | awk '/haproxy-/{print $1}'")
        amp_haproxy_unit = octavia.run_command_on_amphora(
            command=amp_haproxy_unit_command,
            lb_id=self.loadbalancer_stack.loadbalancer_id,
            lb_fip=self.loadbalancer_stack.floating_ip_address)
        LOG.info(f'The amp_haproxy_unit is {amp_haproxy_unit}')

        octavia.run_command_on_amphora(
            command=f'systemctl stop {amp_haproxy_unit}',
            lb_id=self.loadbalancer_stack.loadbalancer_id,
            lb_fip=self.loadbalancer_stack.floating_ip_address,
            sudo=True)

        self._wait_for_failover_and_test_functionality()

    def _skip_if_not_active_standby(self):
        """Skip the test if Octavia doesn't use Active/standby topology
        """
        if len(octavia.list_amphorae(
                self.loadbalancer_stack.loadbalancer_id)) < 2:
            skipping_stmt = 'Skipping the test as it requires ' \
                            'Active/standby topology.'
            LOG.info(skipping_stmt)
            self.skipTest(skipping_stmt)

    def _wait_for_failover_and_test_functionality(self):
        """Wait for failover to end and test Octavia functionality"""

        self.loadbalancer_stack.wait_for_update_loadbalancer()
        self.loadbalancer_stack.wait_for_active_loadbalancer()

        LOG.debug(f'Load Balancer {self.loadbalancer_stack.loadbalancer_id} is'
                  f' ACTIVE')

        # Wait for Octavia objects' provisioning status to be ACTIVE
        self.listener_stack.wait_for_active_members()

        # For 5 minutes we ignore specific exceptions as we know
        # that Octavia resources are being reprovisioned (amphora during a
        # failover)
        for attempt in tobiko.retry(timeout=300.):
            try:
                octavia.check_members_balanced(
                    pool_id=self.listener_stack.pool_id,
                    ip_address=self.loadbalancer_stack.floating_ip_address,
                    lb_algorithm=self.listener_stack.lb_algorithm,
                    protocol=self.listener_stack.lb_protocol,
                    port=self.listener_stack.lb_port)
                break
            except octavia.RoundRobinException:
                LOG.exception(f"Traffic didn't reach all members after "
                              f"#{attempt.number} attempts and "
                              f"{attempt.elapsed_time} seconds")
                if attempt.is_last:
                    raise



# Copyright 2018 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing
import unittest
from unittest import mock

import tobiko
from tobiko.tests import unit


class TestCaseTest(unittest.TestCase):

    def setUp(self):
        super(TestCaseTest, self).setUp()
        self.addCleanup(self._pop_inner_test_cases)

    def _pop_inner_test_cases(self):
        case = tobiko.get_test_case()
        while case is not self:
            tobiko.pop_test_case()
            case = tobiko.get_test_case()

    def test_get_test_case(self):
        result = tobiko.get_test_case()
        self.assertIs(self, result)

    def test_get_test_case_out_of_context(self):
        manager = tobiko.TestCaseManager()
        case = tobiko.get_test_case(manager=manager)
        self.assertIsInstance(case, unittest.TestCase)
        self.assertEqual('tobiko.common._case.DummyTestCase.runTest',
                         case.id())

    def test_push_test_case(self):

        class InnerTest(unittest.TestCase):

            def runTest(self):
                pass

        inner_case = InnerTest()
        tobiko.push_test_case(inner_case)
        self.assertIs(inner_case, tobiko.get_test_case())

    def test_pop_test_case(self):

        class InnerTest(unittest.TestCase):

            def runTest(self):
                pass

        inner_case = InnerTest()
        tobiko.push_test_case(inner_case)

        result = tobiko.pop_test_case()

        self.assertIs(inner_case, result)
        self.assertIs(self, tobiko.get_test_case())

    def test_add_cleanup(self,
                         *args,
                         error: Exception = None,
                         failure: str = None,
                         **kwargs):

        mock_func = mock.Mock()

        class InnerTest(unittest.TestCase):
            def runTest(self):
                tobiko.add_cleanup(mock_func, *args, **kwargs)
                if error is not None:
                    raise error
                if failure is not None:
                    self.fail(failure)

        inner_case = InnerTest()
        mock_func.assert_not_called()
        check = (error is None and failure is None)
        result = tobiko.run_test(case=inner_case,
                                 check=check)
        self.assertEqual(1, result.testsRun)
        mock_func.assert_called_once_with(*args, **kwargs)

        if error is not None:
            self.assertEqual(1, len(result.errors))
            for _error in result.errors:
                self.assertIs(inner_case, _error[0])
                self.assertIn(str(error), _error[1])
        else:
            self.assertEqual([], result.errors)

        if failure is not None:
            self.assertEqual(1, len(result.failures))
            for _failure in result.failures:
                self.assertIs(inner_case, _failure[0])
                self.assertIn(failure, _failure[1])
        else:
            self.assertEqual([], result.failures)

    def test_add_cleanup_with_args(self):
        self.test_add_cleanup(1, 2, a='a', b='b')

    def test_add_cleanup_with_error(self):
        self.test_add_cleanup(error=RuntimeError('some error'))

    def test_add_cleanup_with_failure(self):
        self.test_add_cleanup(failure='some_failure')


class TestFail(unit.TobikoUnitTest):

    def test_fail(self, cause: Exception = None):
        ex = self.assertRaises(tobiko.FailureException, tobiko.fail,
                               'some_reason', cause=cause)
        self.assertEqual('some_reason', str(ex))
        self.assertIs(cause, ex.__cause__)

    def test_fail_with_cause(self):
        self.test_fail(cause=RuntimeError())


class SubtestTest(unittest.TestCase):

    def test_sub_test(self):
        self.assertIs(self, tobiko.get_test_case())
        for item in range(10):
            with self.subTest(f"case-{item}"):
                self.assertIs(self, tobiko.get_test_case())
                self.assertTrue(tobiko.get_sub_test_id().startswith(self.id()))
                self.assertIn(f"case-{item}", tobiko.get_sub_test_id())


class NestedTest(unittest.TestCase):

    executed_id: typing.Optional[str] = None

    def setUp(self) -> None:
        if tobiko.get_parent_test_case() is None:
            self.skipTest('not running as nested test case')

    def test_run_test(self):
        parent = tobiko.get_parent_test_case()
        self.assertIsInstance(parent, unittest.TestCase)
        self.executed_id = self.id()

    def test_run_test_error(self):
        self.executed_id = self.id()
        raise RuntimeError('Planned error')

    def test_run_test_fail(self):
        self.executed_id = self.id()
        self.fail('Planned failure')


class RunTestTest(unittest.TestCase):

    def test_run_test(self):
        nested = NestedTest(methodName='test_run_test')
        result = tobiko.run_test(case=nested)
        self.assertIsInstance(result, unittest.TestResult)
        self.assertEqual(1, result.testsRun)
        self.assertEqual(nested.id(), nested.executed_id)
        self.assertEqual([], result.errors)
        self.assertEqual([], result.failures)

    def test_run_test_error(self):
        nested = NestedTest(methodName='test_run_test_error')

        class ParentTest(unittest.TestCase):
            # pylint: disable=attribute-defined-outside-init
            def runTest(self):
                self.result = tobiko.run_test(case=nested)

        parent = ParentTest()
        result = tobiko.run_test(case=parent, check=False)
        self.assertIsInstance(result, unittest.TestResult)
        self.assertIsInstance(parent.result, unittest.TestResult)
        self.assertEqual(2, result.testsRun)
        self.assertEqual(1, parent.result.testsRun)
        self.assertEqual(nested.id(), nested.executed_id)
        self.assertEqual(nested, parent.result.errors[0][0])
        self.assertEqual(nested, result.errors[0][0])
        self.assertEqual([], result.failures)
        self.assertEqual([], parent.result.failures)

    def test_run_test_fail(self):
        nested = NestedTest(methodName='test_run_test_fail')

        class ParentTest(unittest.TestCase):
            # pylint: disable=attribute-defined-outside-init
            def runTest(self):
                self.result = tobiko.run_test(case=nested)

        parent = ParentTest()
        result = tobiko.run_test(case=parent, check=False)
        self.assertIsInstance(result, unittest.TestResult)
        self.assertIsInstance(parent.result, unittest.TestResult)
        self.assertEqual(2, result.testsRun)
        self.assertEqual(1, parent.result.testsRun)
        self.assertEqual(nested.id(), nested.executed_id)
        self.assertEqual([], result.errors)
        self.assertEqual([], parent.result.errors)
        self.assertEqual(nested, parent.result.failures[0][0])
        self.assertEqual(nested, result.failures[0][0])



# Copyright 2021 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import json
import typing

from oslo_log import log

import tobiko
from tobiko.openstack import keystone
from tobiko.openstack.neutron import _client


LOG = log.getLogger(__name__)


def get_neutron_quota_set(project: keystone.ProjectType = None,
                          client: _client.NeutronClientType = None,
                          detail=False,
                          **params):
    client = _client.neutron_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.httpclient.session)
    if detail:
        return client.show_quota_details(project, **params)['quota']
    else:
        return client.show_quota(project, **params)['quota']


def set_neutron_quota_set(project: keystone.ProjectType = None,
                          client: _client.NeutronClientType = None,
                          **quota):
    client = _client.neutron_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.httpclient.session)
    return client.update_quota(project, body={'quota': quota})['quota']


def ensure_neutron_quota_limits(project: keystone.ProjectType = None,
                                client: _client.NeutronClientType = None,
                                retry_timeout: tobiko.Seconds = None,
                                retry_interval: tobiko.Seconds = None,
                                **required_quotas: int) -> None:
    if not required_quotas:
        return
    client = _client.neutron_client(client)
    project = keystone.get_project_id(project=project,
                                      session=client.httpclient.session)
    for attempt in tobiko.retry(timeout=retry_timeout,
                                interval=retry_interval,
                                default_timeout=60.,
                                default_interval=3.):
        actual_limits, expected_limits = get_neutron_quota_limits_increase(
            project=project, client=client, extra_increase=10//attempt.number,
            **required_quotas)
        if expected_limits:
            if attempt.is_last:
                raise EnsureNeutronQuotaLimitsError(
                    project=project,
                    actual_limits=actual_limits,
                    expected_limits=expected_limits)
            LOG.info(f"Increase Neutron quota limit(s) (project={project}): "
                     f"{actual_limits} -> {expected_limits}...")
            try:
                set_neutron_quota_set(project=project, client=client,
                                      **expected_limits)
            except Exception as ex:
                if attempt.is_last:
                    raise
                LOG.exception("Unable to ensure Neutron quota set limits: "
                              f"{expected_limits}: {ex}")
        else:
            LOG.debug(f"Required quota limits are OK: {required_quotas}")
            break
    else:
        raise RuntimeError("Broken retry loop")


class EnsureNeutronQuotaLimitsError(tobiko.TobikoException):
    message = ("Neutron quota limits lower than "
               "expected (project={project}): "
               "{actual_limits} != {expected_limits}")


def get_neutron_quota_limits_increase(
        project: keystone.ProjectType = None,
        client: _client.NeutronClientType = None,
        extra_increase=0,
        **required_quotas: int) \
        -> typing.Tuple[typing.Dict[str, int],
                        typing.Dict[str, int]]:
    quota_set = get_neutron_quota_set(project=project,
                                      client=client,
                                      detail=True)
    LOG.debug("Got quota set:\n"
              f"{json.dumps(quota_set, indent=4, sort_keys=True)}")
    actual_limits: typing.Dict[str, int] = {}
    expected_limits: typing.Dict[str, int] = {}
    for name, needed in required_quotas.items():
        quota = quota_set[name]
        limit = int(quota['limit'])
        if limit >= 0:
            used = max(0, int(quota['used']))
            reserved = max(0, int(quota['reserved']))
            required_limit = used + reserved + needed
            if required_limit > limit:
                actual_limits[name] = limit
                expected_limits[name] = required_limit + extra_increase
    return actual_limits, expected_limits



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import abc
import typing

import netaddr

import tobiko
from tobiko import config
from tobiko.shell import ping
from tobiko.shell import ssh
from tobiko.openstack import neutron
from tobiko.openstack.stacks import _cirros
from tobiko.openstack.stacks import _neutron
from tobiko.openstack.stacks import _nova


CONF = config.CONF


class VlanNetworkStackFixture(_neutron.NetworkStackFixture):
    pass


class VlanProxyServerStackFixture(_cirros.CirrosServerStackFixture):
    network_stack = tobiko.required_fixture(VlanNetworkStackFixture)


class VlanServerStackFixture(_nova.ServerStackFixture, abc.ABC):

    @property
    def has_vlan(self) -> bool:
        return neutron.has_networking_extensions('trunk')

    #: stack with the newtwork where the trunk support is attached
    vlan_network_stack = tobiko.required_fixture(VlanNetworkStackFixture)

    @property
    def vlan_id(self) -> int:
        return CONF.tobiko.neutron.vlan_id

    @property
    def vlan_network(self) -> str:
        return self.vlan_network_stack.network_id

    @property
    def vlan_fixed_ipv4(self):
        return self.find_vlan_fixed_ip(ip_version=4)

    @property
    def vlan_fixed_ipv6(self):
        return self.find_vlan_fixed_ip(ip_version=6)

    def find_vlan_fixed_ip(self,
                           ip_version: int = None,
                           unique=False) -> netaddr.IPAddress:
        vlan_ips = self.list_vlan_fixed_ips(ip_version=ip_version)
        if unique:
            return vlan_ips.unique
        else:
            return vlan_ips.first

    def list_vlan_fixed_ips(self,
                            ip_version: int = None) \
            -> tobiko.Selection[netaddr.IPAddress]:
        fixed_ips = tobiko.Selection[netaddr.IPAddress]()
        if self.vlan_fixed_ips:
            fixed_ips.extend(netaddr.IPAddress(fixed_ip['ip_address'])
                             for fixed_ip in self.vlan_fixed_ips)
            if ip_version is not None and fixed_ips:
                fixed_ips = fixed_ips.with_attributes(version=ip_version)
        return fixed_ips

    @property
    def vlan_ssh_proxy_client(self) -> ssh.SSHClientType:
        return tobiko.setup_fixture(VlanProxyServerStackFixture).ssh_client

    def assert_vlan_is_reachable(self,
                                 ip_version: int = None,
                                 timeout: tobiko.Seconds = None,
                                 ssh_client: ssh.SSHClientType = None):
        fixed_ips = self.list_vlan_fixed_ips(ip_version=ip_version)
        if fixed_ips:
            if timeout is None:
                timeout = self.is_reachable_timeout
            if ssh_client is None:
                ssh_client = self.vlan_ssh_proxy_client
            ping.assert_reachable_hosts(fixed_ips,
                                        ssh_client=ssh_client,
                                        timeout=timeout)
        else:
            tobiko.fail(f'Server {self.stack_name} has any IP on VLAN port')

    def assert_vlan_is_unreachable(self,
                                   ip_version: int = None,
                                   timeout: tobiko.Seconds = None,
                                   ssh_client: ssh.SSHClientType = None):
        fixed_ips = self.list_vlan_fixed_ips(ip_version=ip_version)
        if fixed_ips:
            if ssh_client is None:
                ssh_client = self.vlan_ssh_proxy_client
            ping.assert_unreachable_hosts(fixed_ips,
                                          ssh_client=ssh_client,
                                          timeout=timeout)
        else:
            tobiko.fail(f'Server {self.stack_name} has any IP on VLAN port')

    @property
    def neutron_required_quota_set(self) -> typing.Dict[str, int]:
        requirements = super().neutron_required_quota_set
        if self.has_vlan:
            requirements['trunk'] += 1
            requirements['port'] += 1
        return requirements
